import os
import time
import torch
import random
import numpy as np

from kevin_toolbox.machine_learning.dataset.face import verification
from kevin_toolbox.data_flow.core.reader import Unified_Reader_Base


def get_fail_case_from_iterator(**kwargs):
    """
        接受数据集的 iterator，然后在给定的 threshold 下挑出 false positive 和 false negative cases，
            当 fail cases 的数量达到目标值时，停止并返回

        选取 fail cases 相关的参数：
            threshold：              <float> 阈值
                                        作为选取 fail cases 的标准是：
                                            false positive：     score > threshold and label == False
                                            false negative：     score < threshold and label == True
            lower_bound：            <integer> 至少选取多少个 false positive / false negative cases
                                        默认为None
            upper_bound：            <integer> 至多选取多少个 false positive / false negative cases
                                        默认为None，亦即只要满足下界即可

        多进程相关的参数
            world_size:             <integer>
            rank:                   <integer>
                        （设定后，将只统计第 n 个数据集，其中 n 满足 n % world_size == rank）

        控制输出相关的参数：
            verbose：                <boolean> 是否打印过程信息

        返回：
            fp_db：                  <dict> 过滤出来的 false positive cases 部分的数据集
            fn_db：                  <dict> 过滤出来的 false negative cases 部分的数据集
    """
    # 默认参数
    paras = {
        # 数据集相关
        "iterator": None,
        # 选取 fail cases 相关
        "threshold": None,
        "lower_bound": None,
        "upper_bound": None,
        # 多进程相关
        "world_size": 1,
        "rank": 0,
        # 控制输出
        "verbose": True,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    # 数据集相关
    assert paras["iterator"] is not None
    # 选取 fail cases 相关
    assert paras["threshold"] is not None
    lower_bound = paras["lower_bound"]
    upper_bound = paras["upper_bound"]
    assert lower_bound is None or (
            isinstance(lower_bound, (int,)) and
            lower_bound > 0)
    assert upper_bound is None or (
            isinstance(upper_bound, (int,)) and
            (lower_bound is None or upper_bound >= lower_bound))
    # 多进程相关
    assert isinstance(paras["world_size"], (int,)) and paras["world_size"] > 0
    assert isinstance(paras["rank"], (int,)) and 0 <= paras["rank"] < paras["world_size"]

    """
    选取
    """

    count = 0
    beg_time = time.perf_counter()
    index_ls = [i for i in range(len(paras["iterator"])) if i % paras["world_size"] == paras["rank"]]
    index_ls = random.sample(index_ls, k=len(index_ls))
    fp_db, fn_db = dict(), dict()
    for i in index_ls:
        #
        db = paras["iterator"][i]
        # false positive cases
        cond_fp = ((db["scores"] > paras["threshold"]) & (db["labels"] == 0)).reshape(-1)
        # false negative cases
        cond_fn = ((db["scores"] < paras["threshold"]) & (db["labels"] == 1)).reshape(-1)
        for key, value in db.items():
            if key in fp_db:
                fp_db[key] = np.concatenate([fp_db[key], db[key][cond_fp]])
                fn_db[key] = np.concatenate([fn_db[key], db[key][cond_fn]])
            else:
                fp_db[key] = db[key][cond_fp]
                fn_db[key] = db[key][cond_fn]
        #
        if lower_bound is None:
            rate_n = 0
        else:
            rate_n = min(len(fp_db.get("labels", [])), len(fn_db.get("labels", []))) / lower_bound
            if rate_n > 1:
                break
        #
        if paras["verbose"]:
            count += 1
            rate = max(count / len(index_ls), rate_n)
            time_gap = time.perf_counter() - beg_time
            print(f"--world_size {paras['world_size']} --rank {paras['rank']}\n"
                  f"\t progress rate: {rate * 100}% \t eta: {time_gap / rate / 60 / 60 * (1 - rate)} hours")

    #
    if upper_bound is not None:
        for db in [fp_db, fn_db]:
            db_size = len(list(db.values())[0])
            if db_size > upper_bound:
                index = random.sample(range(db_size), upper_bound)
                for key, value in db.items():
                    db[key] = value[index]
            else:
                continue

    return fp_db, fn_db



import numpy as np
from .computational_tree import Node


def split_whole_into_crops(**kwargs):
    """
        将变量 whole 按照 box_ls 指示的位置，以行优先的内存顺序，拆解成多个 crop
            系 concat_crops_into_whole() 的逆变换

        参数：
            whole:          <np.array/tensor>
            box_ls:         <list of np.arrays>
            beg_axis:       <integer> 要对 x 进行分割的轴
            computational_tree: <Node> 计算图
            return_details: <boolean> 是否以详细信息的形式返回结果
                                默认为 False，此时返回：
                                    crop_ls:  <list of np.array/tensor> 分割结果
                                当设置为 True，将返回一个 dict：
                                    details = dict(
                                        whole = <np.array/tensor>,  # 对 crop_ls 进行合并后的结果
                                        box_ls = <list of np.arrays>,  # 按照 内存顺序 对 box_ls 进行排序后的结果
                                        crop_ls = <list of np.array/tensor>,  # 按照 内存顺序 对 crop_ls 进行排序后的结果
                                        beg_axis = beg_axis,  # 对应与输入的 beg_axis
                                        computational_tree = <Node>,  # 计算图
                                    )
                （以上参数的详细介绍参见 concat_crops_in_memory_order()）
        返回：
            crop_ls 或者 details
    """
    # 默认参数
    paras = {
        # 必要参数
        "whole": None,
        "box_ls": None,
        #
        "beg_axis": 0,
        "computational_tree": None,
        "return_details": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert paras["whole"] is not None
    whole = paras["whole"]
    #
    assert isinstance(paras["box_ls"], (list,)) and paras["box_ls"][0].ndim == 2 and paras["box_ls"][0].shape[1] == 2
    box_ls = paras["box_ls"]
    #
    assert isinstance(paras["beg_axis"], (int,)) and 0 <= paras["beg_axis"] < whole.ndim
    beg_axis = paras["beg_axis"]
    # 构建计算图
    if paras["computational_tree"] is None:
        tree = Node(box_ls=box_ls)
        tree.build_tree()
        tree.init_tree()
    else:
        tree = paras["computational_tree"]
    assert isinstance(tree, (Node,))

    tree.var = whole
    tree.split(beg_axis=beg_axis)

    # 按行优先进行多级排序
    sorted_node_ls = sorted(tree.get_leaf_nodes(), key=lambda x: x.details["box_ls"][0][0].tolist())
    sorted_crop_ls = [node.var for node in sorted_node_ls]
    sorted_box_ls = [node.details["box_ls"][0] for node in sorted_node_ls]

    if paras["return_details"]:
        details = dict(
            whole=whole,
            crop_ls=sorted_crop_ls,
            box_ls=sorted_box_ls,
            beg_axis=beg_axis,
            computational_tree=tree,
        )
        return details
    else:
        return crop_ls



from kevin_toolbox.data_flow.core.reader import UReader, Unified_Reader_Base
from .cal_cfm import cal_cfm
from .merge_cfm_ls import merge_cfm_ls
from .convert import convert_to_numpy


def cal_cfm_iteratively_by_chunk(scores, labels, chunk_step=None, to_numpy=True, decimals_of_scores=None, **kwargs):
    """
        迭代地从 scores, labels 和取出一部分来通过 cal_cfm() 分别计算出 confusion_matrices，
        然后再使用 merge_cfm_ls() 对结果进行合并。
            避免直接对全部进行计算造成对内存占用过大。

        参数（详细介绍参见 cal_cfm()）：
            scores:             表示模型将样本预测为正样本的置信度
            labels:             描述样本的真实标签，其中1表示正样本，0表示负样本。
                注意： scores 和 labels 支持变量的形式，和使用 Unified_Reader_Base 包裹的形式

        可选参数：
            chunk_step:         integer，每个分块的大小
                                    默认为 None，表示将所有输入视为一个分块，此时该函数退化为 cal_cfm()
            to_numpy:           boolean，是否将结果从tensor转换为numpy
                                    默认为 True
            decimals_of_scores：     integer，scores 要保留的小数位数（直接截取而非四舍五入）
                                    默认为 None，表示不对 scores 进行操作。
                                    实际上，scores的取值有非常多，
                                    为了避免 thresholds 过多，可以通过限制 scores 的小数位，来对结果进行进一步合并和去重，
                                    从而减少内存占用。

        返回（详细介绍参见 cal_cfm()）：
            confusion_matrices： 描述了不同 threshold 取值下，二分类混淆矩阵中的各项 TP、FN、FP、TN 的取值
    """
    scores = scores if isinstance(scores, (Unified_Reader_Base,)) else UReader(var=scores)
    labels = labels if isinstance(labels, (Unified_Reader_Base,)) else UReader(var=labels)
    assert isinstance(scores, (Unified_Reader_Base,)) and isinstance(labels, (Unified_Reader_Base,)), \
        Exception(f"Error: scores and labels need to be wrapped with Unified_Reader_Base!")
    assert len(scores) == len(labels) > 0
    chunk_step = len(scores) if chunk_step is None else chunk_step
    assert isinstance(chunk_step, (int,)) and 0 < chunk_step <= len(scores), \
        Exception(f"Error: chunk_step({chunk_step}) should be between 1 and len(scores)({len(scores)}),"
                  f" or use the default value of None!")

    chunk_nums = (len(scores) - 1) // chunk_step + 1
    cfm_ls = []
    for i in range(chunk_nums):
        beg = i * chunk_step
        end = beg + chunk_step
        cfm = cal_cfm(scores.read(beg, end), labels.read(beg, end),
                      to_numpy=False, decimals_of_scores=decimals_of_scores)
        cfm_ls.append(cfm)
        if len(cfm_ls) >= 3:  # 计算多少个就合并一次
            cfm_ls = [merge_cfm_ls(cfm_ls, to_numpy=False)]

    if len(cfm_ls) > 1:
        cfm = merge_cfm_ls(cfm_ls, to_numpy=False)
    else:
        cfm = cfm_ls[0]

    if to_numpy:
        cfm = convert_to_numpy(cfm, decimals_of_scores)

    return cfm



import numpy as np


def convert_boxes_from_coord_to_grid_index(**kwargs):
    """
        对输入的 boxes，进行 实数坐标 与 网格点序号坐标 之间的坐标转换
            注意：
            - 这种转换可能是可逆的， 实数坐标 ==> 网格点序号坐标 的转换一般会使得 box 的实际范围扩大。
            - 特别地，当使用 grid_coverage_mode=closed 模式，并配合从 boxes 中获取的坐标（可以通过for_boxes.get_ticks()获取）时，
                转换是完全可逆的。
            - 网格点的 index 包头不包尾， beg, end = 0, 1 表示 0 号网格。

        参数：
            boxes:          <3 axis np.array> 需要转换的 box
                                shape [batch_size, 2, dimensions]，各个维度的意义为：
                                    batch_size： 有多少个 box
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
            settings_for_grid:      <dict of paras> 用于设定网格位置、范围的参数列表。
                                目前支持两种格点覆盖模式 mode：
                                    mode:     <string> 格点覆盖模式
                                        支持以下两种模式：
                                            "open"：     开放式，将构建一个覆盖整个空间的格点阵列。
                                            "closed"：   封闭式，仅在指定范围内构建格点阵列。对于超出范围外的坐标，将投影到格点阵列的边界上。
                                在不同的格点覆盖模式 mode 下，有不同的设置方式。
                                目前支持以下三种方式：
                                    mode=open，以 grid_size 为基准
                                        grid_size:      <list/integer/float> 各个维度上网格的大小
                                                            设置为单个 integer 时，默认所有维度使用同一大小的网格划分
                                        offset：         <list/integer/float> 网格点的原点相对于原始坐标的偏移量
                                                            默认为 [0,...]，无偏移
                                                            设置为单个 integer 时，默认所有维度使用同一大小的 offset
                                            例子：
                                                当 grid_size=[1,5] , offset=[3,1]，
                                                表示以 coord=(3,1) 为原点，对维度dim=0以size=1划分网格，对dim=1以size=5划分网格。
                                    mode=closed，以 ticks 为基准
                                        ticks：          <list of np.array> 在各个维度下，网格点的划分坐标
                                                            ticks[i][0] 就是网格的原点坐标，与上面的 offset 相同
                                    mode=closed，以 grid_size 为基准
                                        grid_size:      <list of np.array/list> 在各个维度下，网格点的一系列划分大小
                                        offset：         <list/integer>
                                            函数将会首先把 grid_size 和 offset 转换为对应的 ticks，然后再按照 ticks 执行划分。
            reverse:        <boolean> 决定转换的方向
                                默认为 False，此时为 coord ==> grid_index
                                True 时为 grid_index ==> coord
    """

    # 默认参数
    paras = {
        # 必要参数
        "boxes": None,
        # 格点划分方式
        "grid_coverage_mode": "open",
        "settings_for_grid": None,
        #
        "reverse": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    # boxes
    paras["boxes"] = np.asarray(paras["boxes"], dtype=np.float32)
    assert paras["boxes"].ndim == 3 and paras["boxes"].shape[1] == 2
    boxes = np.sort(paras["boxes"], axis=1)
    # grid
    assert isinstance(paras["settings_for_grid"], (dict,))
    settings = paras["settings_for_grid"]
    assert settings["mode"] in ["open", "closed"]
    mode = settings["mode"]
    # 基础检查
    if "offset" not in settings:
        settings["offset"] = 0
    for key in settings.keys():
        if key in ["grid_size", "offset"]:
            if isinstance(settings[key], (int, float,)):
                settings[key] = [settings[key]] * boxes.shape[-1]
        #
        if key not in ["mode"]:
            assert isinstance(settings[key], (list,)) and boxes.shape[-1] == len(settings[key])
    # 结合 mode 进行检查
    if mode == "open":
        for key in ["grid_size", "offset"]:
            assert key in settings and isinstance(settings[key][0], (int, float,))
    else:
        if "ticks" in settings:
            # convert to np.array
            for i, tk_ls in enumerate(settings["ticks"]):
                settings["ticks"][i] = np.asarray(tk_ls)
            # from ticks get offset
            settings["offset"] = []
            for tk_ls in settings["ticks"]:
                settings["offset"].append(tk_ls[0])
            # from ticks get grid_size
            settings["grid_size"] = []
            for ot, tk_ls in zip(settings["offset"], settings["ticks"]):
                temp = tk_ls - ot
                gd = temp[1:] - temp[:-1]
                settings["grid_size"].append(gd)
        else:
            for key in ["grid_size", "offset"]:
                assert key in settings
            assert not isinstance(settings["grid_size"][0], (int, float,))
            # from grid_size,offset get ticks
            settings["ticks"] = []
            for ot, gd_ls in zip(settings["offset"], settings["grid_size"]):
                tk_ls = np.zeros(shape=len(gd_ls) + 1)
                tk_ls[0] = ot
                tk_ls[1:] = gd_ls
                tk_ls = np.cumsum(tk_ls)
                settings["ticks"].append(tk_ls)

    if mode == "open":
        grid_size, offset = settings["grid_size"], np.asarray(settings["offset"])
        if not paras["reverse"]:
            boxes = boxes - offset
            boxes = boxes / grid_size
            boxes[:, 0, :] = np.floor(boxes[:, 0, :])
            boxes[:, 1, :] = np.ceil(boxes[:, 1, :])
            boxes = boxes.astype(np.int32)
        else:
            boxes = boxes * grid_size
            boxes = boxes + offset
    else:
        grid_size, ticks = settings["grid_size"], settings["ticks"]
        if not paras["reverse"]:
            temp_ls = []
            for dim in range(boxes.shape[-1]):
                temp = boxes[..., [dim]] - ticks[dim].reshape(1, -1)
                temp[..., :-1] = temp[..., :-1] / grid_size[dim]
                #
                temp[:, 0, :] = np.floor(temp[:, 0, :])
                temp[:, 1, :] = np.ceil(temp[:, 1, :])
                #
                temp = np.maximum(np.minimum(temp, 1), 0)
                temp = np.sum(temp, axis=-1, keepdims=True)
                temp_ls.append(temp)
            boxes = np.concatenate(temp_ls, axis=-1)
            boxes = boxes.astype(np.int32)
        else:
            temp = np.zeros_like(boxes)
            for i in range(boxes.shape[0]):
                for j in range(boxes.shape[1]):
                    for dim in range(boxes.shape[-1]):
                        temp[i, j, dim] = ticks[dim][int(boxes[i, j, dim])]
            boxes = temp

    return boxes


if __name__ == '__main__':
    settings_for_grid = dict(mode="open", grid_size=2, offset=1)
    grids = convert_boxes_from_coord_to_grid_index(boxes=np.array([[[1, 3, 4, ],
                                                                    [2, 1, -6]]]),
                                                   settings_for_grid=settings_for_grid)
    print(grids)
    boxes_ = convert_boxes_from_coord_to_grid_index(boxes=grids, settings_for_grid=settings_for_grid,
                                                    reverse=True, )
    print(boxes_)



import os
import pickle
from kevin_toolbox.computer_science.data_structure import Executor


class Egg:
    """
        一个以 executor 为“蛋壳”，包裹着内容 var 的“鸡蛋”。
            在 pickle 模块对其进行“孵化”时，将会首先“执行”“蛋壳”，然后再获取其内容。
    """
    def __init__(self, executor, var=None):
        self.executor = executor
        self.var = var

    def __reduce__(self):
        """
            该函数系 pickle 供给用户用于指定反序列化的，将在 pickle 解释该对象时被自动执行
                https://docs.python.org/3/library/pickle.html
        """
        assert self.executor is not None

        func, args_, kwargs_ = self.executor.parse()

        return (eval, ("(func(*args_, **kwargs_), var)[-1]",
                       {"func": func, "args_": args_, "kwargs_": kwargs_, "var": self.var}),)


def dump_into_pickle_with_executor_attached(var, executor, output_path):
    """
        将输入的变量 var 打包成一个附带着 executor 的 pickle 文件

        何为附带着 executor 的 pickle 文件，它具有什么特性？
            - 读取/解释该 pickle 文件时，将自动触发并执行其附带的 executor。
                    executor 系 kevin_toolbox.computer_science.data_structure.Executor 的一个实例，
                    它允许你以静态方式描述并保存一个执行过程，然后在你需要的地方再进行调用。
            - 从该 pickle 文件读取得到的结果，与普通地将 var 保存成 pickle 文件读取得到的结果完全一致。

            警告！ 该函数并不对 executor 的执行内容负责，因此可能产生对执行环境危害性极大的 pickle 文件！

            比如：
                import os
                from kevin_toolbox.computer_science.data_structure import Executor

                # 创建一个 executor，执行它会调用命令在用户目录下创建一个文件（这个命令也可以换成任何更加危险的命令，小朋友不要学坏哦）
                executor = Executor(func=os.system, args=["touch ~/test.txt"])
                build_pickle_to_run(executor, output_path="~/egg")

            然后在另一个进程中尝试读取该 pickle 文件：
                import pickle

                pickle.load(open("~/egg","rb"))
            它将会自动执行前面的 executor。

        参数：
            var:                需要被打包的变量
            executor:           <Executor> 需要打包的执行器
            output_path:        <str> 输出 pickle 文件的保存路径
    """
    assert isinstance(executor, (Executor,))

    egg = Egg(executor=executor, var=var)

    os.makedirs(os.path.split(output_path)[0], exist_ok=True)
    with open(output_path, "wb") as f:
        pickle.dump(egg, f)



from kevin_toolbox.computer_science.algorithm.for_nested_dict_list import traverse


def get_leaf_nodes(var):
    """
        获取嵌套字典列表 var 中所有叶节点
            以列表 [(name,value), ...] 形式返回，其中名字 name 的解释方式参考 get_value_by_name() 介绍

        参数：
            var:                待处理数据
                                    当 var 不是 dict 或者 list 时，返回空列表
    """
    res = []

    def func(_, idx, v):
        nonlocal res
        if not isinstance(v, (list, dict,)):
            res.append((idx, v))
        return False

    traverse(var=var, match_cond=func, action_mode="skip", b_use_name_as_idx=True)
    return res



# 针对唯一标识符的单例模式
class Singleton_for_uid:
    """
        针对唯一标识符的单例模式
        对于一个给定的 uid 只会进行一次实例的生成，以后再对同一个 uid 调用 Singleton_for_uid() 进行实例生成，返回的只有以前生成的实例。
    """
    __instances = dict()  # {uid:<instance>, ...}
    __counter = 0
    class_name = ""

    def __new__(cls, *args, **kwargs):
        """
            __new__函数返回的实例，将作为self参数被传入到__init__函数。
                如果__new__函数返回一个已经存在的实例（不论是哪个类的），__init__还是会被调用的，所以要特别注意__init__中对变量的赋值。
        """
        uid = kwargs.get("uid", None)
        exist_ok = kwargs.get("exist_ok", True)
        assert isinstance(uid, (str, type(None),)), \
            TypeError(f"uid should be string or None, but get a {type(uid)}!")

        if uid in Singleton_for_uid.__instances:
            if exist_ok:
                # 返回已有实例，__init__不会被调用
                return Singleton_for_uid.__instances.get(uid)
            else:
                raise Exception(f"uid {uid} already exists!")
        else:
            # 传入 __init__
            return super().__new__(cls)

    def __init__(self, *args, **kwargs):
        try:
            getattr(self, "uid")
        except :
            pass
        else:
            return
        uid = kwargs.get("uid", None)
        assert isinstance(uid, (str, type(None),)), \
            TypeError(f"uid should be string or None, but get a {type(uid)}!")

        if uid is None:
            # 生成一个未被注册的 uid
            while True:
                uid = f"{Singleton_for_uid.class_name}_{Singleton_for_uid.__counter}"
                if uid not in Singleton_for_uid.__instances:
                    break
                Singleton_for_uid.__counter += 1
        self.uid = uid
        # 注册
        Singleton_for_uid.__instances[self.uid] = self

    @staticmethod
    def register_out(*args, **kwargs):
        """
            建议先使用本函数，将实例从静态变量__instances中删除，然后再使用 del 来删除实例。
                由于在实现单例模式时，需要使用静态变量来记录实例。
                因此无法直接通过 del 来消除对于某个实例的所有引用，来触发gc垃圾回收机制。

            建议与 del 联用：
                ins_to_del = Singleton_for_uid.register_out( uid=ins_uid )
                del ins_to_del
        """
        uid = kwargs.get('uid', None)
        assert uid in Singleton_for_uid.__instances, \
            Exception(f"uid {uid} not exists!")
        return Singleton_for_uid.__instances.pop(uid)

    @staticmethod
    def get_instances():
        return Singleton_for_uid.__instances


if __name__ == '__main__':
    class Node(Singleton_for_uid):
        Singleton_for_uid.class_name = "Node"

        def __init__(self, *args, **kwargs):
            super(Node, self).__init__(*args, **kwargs)

        def __del__(self):
            print("deleting ", self)


    "删除"
    node = Node(uid="ins_1")
    Node.register_out(uid=node.uid)
    del node
    print(2333)
    # deleting  <__main__.Singleton_for_uid object at 0x7f150c270b50>
    # 2333
    # 在 print 之前成功触发 del

    "测试单例模式"
    node_1 = Node(uid="111")
    node_2 = Node(uid="111")
    node_3 = Node()
    print(id(node_1), id(node_2), id(node_3))
    print(Node.get_instances())
    # 140482309115472 140482309115472 140482309115536
    # {'111': <__main__.Node object at 0x7fc49621ce50>, 'Node_0': <__main__.Node object at 0x7fc49621ce90>}

    node_4 = Node(uid="111", exist_ok=False)



import os
import time
import importlib.util
from kevin_toolbox.developing.decorator import restore_original_work_path

if importlib.util.find_spec("cPickle") is not None:
    # 如果安装有 cPickle，可以更快处理串行化
    import cPickle as pickle
else:
    import pickle


class Cache_Manager_for_Iterator:
    """
        适用于迭代器/生成器的缓存管理器
            通过构建基于磁盘的缓存（首先迭代读取迭代器/生成器的内容，并将这些块 chunk 分别保存为二进制文件）
            以及基于内存的缓存（然后在内存中开辟出一个缓存列表空间，用于保存最常调用的 chunk）
            来加速 迭代器/生成器 内容的读取

        工作流程：
            首先构建基于磁盘的缓存，将分块读取到数据保存为二进制文件；
                相关变量：
                    iterator:       迭代器/生成器
                    folder_path:    保存二进制文件的路径
                                        默认保存在 ./temp/cache_name/ 下
                                        当给定的 folder_path 内容不为空时，将尝试直接构建 file_dict
                    file_dict:      二进制文件的文件名与序号的对应关系表
                                        例如：{ 0: "0.pkl", 1: "1.pkl", 2: "2.pkl", ...}
                                        其中 index 0 对应于文件 ./temp/cache_name/0.pkl
            然后在进行读取时，将先到 cache 中寻找是否已经有需要的 chunk 分块，如果没有则到前面 file_dict 中读取，同时更新 cache
                相关变量：
                    cache_dict：     缓存
                                        例如：{ 1: chunk_var_1, 3: chunk_var_3 }
                                        其中 key 是 chunk 分块的 index，value 是对应的保存在内存中的变量
                    cache_metadata:  缓存的属性数据
                                        包含以下字段，各字段将自动更新
                                        例如：{ 1: {  "last_time": xxx,      # 最近读取时间
                                                     "initial_time": xxx,   # 最初读取时间
                                                     "counts": xxx,         # 读取次数
                                                     },
                                               ... }
                    cache_update_strategy：缓存更新策略
                                        是一个函数，该函数的输入是 cache_metadata ，输出是需要删除的缓存的序号
                                        触发：
                                            在每次出现 cache_dict 无法命中，导致有新的 cache 添加到 cache_dict 时，将会采用该策略进行更新
                                        现有策略：
                                            drop_min_counts:    去除读取次数最小的
                                            drop_min_last_time:    去除最近没有读取的
                                            drop_min_survival_time:    去除生存时间最短的，生成时间 survival_time:=last_time-initial_time
                                        默认使用 drop_min_last_time
                    （cache_dict 的大小受 cache_update_strategy 中的 cache_size 限制，当大小超过限制时，
                    将根据 cache_update_strategy 去除优先级较低的部分来更新缓存）

        支持以下几种方式来：
            以迭代器的形式进行顺序读取
            以指定序号的形式进行随机读取
    """

    def __init__(self, **kwargs):
        """
            设定关键参数
            参数：
                iterator:       迭代器/生成器
                folder_path:    保存二进制文件的路径
                cache_update_strategy：缓存更新策略
            其他参数：
                strict_mode:    禁止同时设置 iterator 和给定一个非空的 folder_path
                                    默认为 True 开启，此时同时设置将报错。
                                    当设置为 False 时，同时设置将以 folder_path 中的二进制文件为准
                del_cache_when_exit:    退出时删除生成的缓存二进制文件
                                    只有在设置了 iterator 的前提下，才会触发。
                                    （对于非本实例生成的文件，比如只给定了非空的 folder_path，不做删除。）
                                    默认为 True 开启。
        """

        # 默认参数
        paras = {
            "iterator": None,
            "folder_path": None,
            "cache_update_strategy": None,
            #
            "strict_mode": True,
            "del_cache_when_exit": True,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        # cache_update_strategy
        if paras["cache_update_strategy"] is None:
            paras["cache_update_strategy"] = lambda x: Strategies.drop_min_last_time(cache_metadata=x,
                                                                                     cache_size_upper_bound=10)
        # 同时非空
        b_folder_not_empty = isinstance(paras["folder_path"], (str,)) and paras[
            "folder_path"] is not None and os.path.exists(paras["folder_path"]) and len(
            os.listdir(paras["folder_path"])) > 0
        if paras["iterator"] is not None and b_folder_not_empty:
            # iterator 非空，folder_path 非空
            if paras["strict_mode"]:
                # 不能同时设置
                raise Exception(f"Error: folder_path and iterator cannot be set at the same time\n"
                                f"iterator {paras['iterator']} is given when "
                                f"there is already content in folder_path {paras['folder_path']}!")
            else:
                # 以 folder_path 为准
                paras["iterator"] = None
        # 同时为空
        if paras["iterator"] is None and paras["folder_path"] is None:
            raise Exception(f"Error: folder_path and iterator cannot be empty at the same time\n"
                            f"both iterator and folder_path are not given!")

        # 构建基于磁盘的缓存
        file_dict = dict()
        if paras["iterator"] is not None:
            # 根据 iterator 生成
            if paras["folder_path"] is None:
                paras["folder_path"] = os.path.join(os.getcwd(), "temp", str(time.time()))
            if not os.path.exists(paras["folder_path"]):
                os.makedirs(paras["folder_path"])
            file_dict = self.generate_chuck_files(paras["iterator"], paras["folder_path"])
        elif b_folder_not_empty:
            # 尝试直接根据已有文件构建 file_dict
            file_dict = self.find_chuck_files(paras["folder_path"])

        self.file_dict = file_dict
        self.paras = paras

        # 初始化基于内存的缓存
        self.cache_dict = dict()
        self.cache_metadata = dict()

        # 记录最后读取的index
        self.index = -1

    # ------------------------------------ 基于磁盘的缓存 ------------------------------------ #

    @staticmethod
    @restore_original_work_path
    def generate_chuck_files(iterator, folder_path):
        """
            构建基于磁盘的缓存
                将 iterator 每次迭代产生的结果以二进制文件的形式（文件名为 {index}.pkl）保存到 folder_path 指向的目录中，
                并将这些文件名保存到索引 file_dict
        """
        os.chdir(folder_path)
        file_dict = dict()
        for i, chuck in enumerate(iterator):
            file_name = f"{i}.pkl"
            with open(file_name, 'wb') as f:
                pickle.dump(chuck, f)
            file_dict[i] = file_name
        return file_dict

    @staticmethod
    @restore_original_work_path
    def find_chuck_files(folder_path):
        """
            建立到磁盘缓存的索引
                从 folder_path 中找到已有的二进制文件，并将这些文件名保存到索引 file_dict
        """
        os.chdir(folder_path)
        file_dict = dict()
        count = 0
        while True:
            file_name = f"{count}.pkl"
            if os.path.isfile(file_name):
                file_dict[count] = file_name
            else:
                break
            count += 1
        return file_dict

    def __read_from_files(self, index):
        """
            从磁盘中读取
                根据 index 到索引 file_dict 中找出对应的文件名，然后读取文件
        """
        file_path = os.path.join(self.paras["folder_path"], self.file_dict[index])
        with open(file_path, 'rb') as f:
            chunk = pickle.load(f)
        return chunk

    # ------------------------------------ 基于内存的缓存 ------------------------------------ #

    def __read_from_cache(self, index):
        """
            从内存中读取
        """
        chunk = self.cache_dict[index]
        # 更新缓存属性
        self.cache_metadata[index]["counts"] += 1
        self.cache_metadata[index]["last_time"] = time.time()
        return chunk

    def __add_to_cache(self, index, chunk):
        """
            添加到内存中
        """
        # 更新缓存
        self.cache_dict[index] = chunk
        # 更新缓存属性
        self.cache_metadata[index] = {
            "last_time": time.time(),  # 最近读取时间
            "initial_time": time.time(),  # 最初读取时间
            "counts": 1,  # 读取次数
        }
        # 依据策略去除优先级较低的缓存
        drop_ls = self.paras["cache_update_strategy"](self.cache_metadata)
        for i in drop_ls:
            self.cache_dict.pop(i)
            self.cache_metadata.pop(i)

    # ------------------------------------ 读取 ------------------------------------ #

    def read(self, index):
        assert 0 <= index < len(self), \
            KeyError(f"Error: index {index} not in [0, {len(self)})")
        if index in self.cache_dict:
            # 直接从内存中读取
            chunk = self.__read_from_cache(index)
        else:
            # 到磁盘读取
            chunk = self.__read_from_files(index)
            # 添加到缓存
            self.__add_to_cache(index, chunk)
        self.index = index
        return chunk

    # ------------------------------------ 其他 ------------------------------------ #

    # 迭代器，通过 next(self) 调用
    def __next__(self):
        if self.index < len(self) - 1:
            return self.read(self.index + 1)
        else:
            self.index = -1
            raise StopIteration

    # 支持 for 循环调用
    def __iter__(self):
        return self

    # 通过 len(self) 调用
    def __len__(self):
        return len(self.file_dict)

    def __del__(self):
        if self.paras["iterator"] is not None and self.paras["del_cache_when_exit"] and self.paras["strict_mode"]:
            # 在 strict_mode 开启，且 iterator 非空的情况下 self.file_dict 中的二进制文件一定是根据 iterator 生成的
            # 删除文件
            pwd_bak = os.getcwd()
            os.chdir(self.paras["folder_path"])
            for key, value in self.file_dict.items():
                os.remove(value)
            os.chdir(pwd_bak)
            # 删除空文件夹
            if not os.listdir(self.paras["folder_path"]):
                os.removedirs(self.paras["folder_path"])


class Strategies:
    """
        现有策略：
                drop_min_counts:    去除读取次数最小的
                drop_min_last_time:    去除最近没有读取的
                drop_min_survival_time:    去除生存时间最短的，生成时间 survival_time:=last_time-initial_time
    """

    @staticmethod
    def drop_min_counts(cache_metadata, cache_size_upper_bound, cache_size_after_drop=None):
        """
            去除读取次数最小的
            参数：
                cache_metadata:  缓存的属性数据
                                    包含以下字段，各字段将自动更新
                                    例如：{ 1: {  "last_time": xxx,      # 最近读取时间
                                                 "initial_time": xxx,   # 最初读取时间
                                                 "counts": xxx,         # 读取次数
                                                 },
                                           ... }
                cache_size_upper_bound： 当 cache_metadata 的大小超过该值时触发更新
                cache_size_after_drop：  更新后 cache_metadata 的目标大小
                                    默认为 cache_size_upper_bound
        """
        if cache_size_upper_bound >= len(cache_metadata):
            return []

        cache_size_after_drop = cache_size_upper_bound if cache_size_after_drop is None else cache_size_after_drop
        # （这里其实可以用最大堆来优化，但是我懒啊）
        drop_ls = [i for i, j in sorted(cache_metadata.items(), key=lambda x: x[1]["counts"])[:-cache_size_after_drop]]
        return drop_ls

    @staticmethod
    def drop_min_last_time(cache_metadata, cache_size_upper_bound, cache_size_after_drop=None):
        """
            去除最近没有读取的
        """
        if cache_size_upper_bound >= len(cache_metadata):
            return []

        cache_size_after_drop = cache_size_upper_bound if cache_size_after_drop is None else cache_size_after_drop
        drop_ls = [i for i, j in
                   sorted(cache_metadata.items(), key=lambda x: x[1]["last_time"])[:-cache_size_after_drop]]
        return drop_ls

    @staticmethod
    def drop_min_survival_time(cache_metadata, cache_size_upper_bound, cache_size_after_drop=None):
        """
            去除生存时间最短的，生成时间 survival_time:=last_time-initial_time
        """
        if cache_size_upper_bound >= len(cache_metadata):
            return []

        cache_size_after_drop = cache_size_upper_bound if cache_size_after_drop is None else cache_size_after_drop
        drop_ls = [i for i, j in sorted(cache_metadata.items(), key=lambda x: x[1]["last_time"] - x[1]["initial_time"])[
                                 :-cache_size_after_drop]]
        return drop_ls


if __name__ == '__main__':
    "测试 Strategies"
    _cache_metadata = {
        1: {
            "last_time": 123,
            "initial_time": 56,
            "counts": 2,
        },
        2: {
            "last_time": 110,
            "initial_time": 0,
            "counts": 4,
        },
        3: {
            "last_time": 126,
            "initial_time": 121,
            "counts": 1,
        },
    }

    print(Strategies.drop_min_counts(_cache_metadata, 2))
    print(Strategies.drop_min_last_time(_cache_metadata, 1))
    print(Strategies.drop_min_survival_time(_cache_metadata, 1))

    "测试 Cache_Manager_for_Iterator"
    cache_manager = Cache_Manager_for_Iterator(iterator=range(10),
                                               del_cache_when_exit=True,
                                               cache_update_strategy=lambda x: Strategies.drop_min_last_time(
                                                   cache_metadata=x,
                                                   cache_size_upper_bound=3))
    print(cache_manager.file_dict)
    print(cache_manager.cache_metadata)
    for i in range(3):
        print(cache_manager.read(0))
    for i in range(2):
        print(cache_manager.read(3))
    for i in range(1):
        print(cache_manager.read(6))
    for i in range(4):
        print(cache_manager.read(9))
    print(cache_manager.cache_metadata)



import numpy as np
import torch
import warnings


def check_consistency(*args, tolerance=1e-7, require_same_shape=True):
    """
        检查 args 中多个 array 之间是否一致

        工作流程：
            本函数会首先将输入的 args 中的所有变量转换为 np.array;
            然后使用 issubclass() 判断转换后得到的变量属于以下哪几种基本类型：
                - 当所有变量都属于 np.number 数值（包含int、float等）或者 np.bool_ 布尔值时，
                    将对变量两两求差，当差值小于给定的容许误差 tolerance 时，视为一致。
                - 当所有变量都属于 np.flexible 可变长度类型（包含string等）或者 np.object 时，
                    将使用==进行比较，当返回值都为 True 时，视为一致。
                - 当变量的基本类型不一致（比如同时有np.number和np.flexible）时，
                    直接判断为不一致。
            numpy 中基本类型之间的继承关系参见： https://numpy.org.cn/reference/arrays/scalars.html

        参数：
            tolerance:          <float> 判断 <np.number/np.bool_> 之间是否一致时，的容许误差。
                                    默认为 1e-7。
            require_same_shape: <boolean> 是否强制要求变量的形状一致。
                                    默认为 True，
                                    当设置为 False 时，不同形状的变量可能因为 numpy 的 broadcast 机制而在比较前自动 reshape 为相同维度，进而可能通过比较。
    """
    assert len(args) >= 2
    assert isinstance(tolerance, (int, float,))

    args = [v.detach().cpu() if torch.is_tensor(v) else v for v in args]
    args = [np.asarray(v) for v in args]

    for v in args[1:]:
        if require_same_shape:
            assert args[0].shape == v.shape
        if issubclass(args[0].dtype.type, (np.number, np.bool_,)):
            # 数字类型
            assert issubclass(v.dtype.type, (np.number, np.bool_,))
            if args[0].size > 0:
                assert np.max(np.abs(args[0] - v)) < tolerance, \
                    f"{args[0]}, {v}, deviation: {np.max(np.abs(args[0] - v))}"
        elif issubclass(args[0].dtype.type, (np.flexible, object,)):
            # 可变长度类型
            assert issubclass(v.dtype.type, (np.flexible, object,))
            assert (args[0] == v).all()
        else:
            raise ValueError



import os
import copy
import numpy as np
import warnings
from kevin_toolbox.data_flow.file.kevin_notation.converter import Converter, CONVERTER_FOR_WRITER
from kevin_toolbox.data_flow.file import kevin_notation

np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)


class Kevin_Notation_Writer:
    """
        遵守 kevin_notation 格式的数据文本写入器（格式要求参见本模块下的 readme）
            支持分批次向文件写入内容
    """

    def __init__(self, **kwargs):
        """
            设定关键参数

            必要参数：
                file_path:          <string> 文件路径
            写入相关参数：
                mode:               <string> 写入模式
                                        支持以下模式：
                                            "w":    从头开始写入
                                            "a":    从末尾续写（要求文件已经具有 metadata）
                paras_for_open:     <paras dict> open() 函数的补充参数（除 mode 以外）
                converter:          <instance of kevin_toolbox.Converter> converter is a dictionary-like data structure
                                            consisting of <string>:<func> pairs，
                                            用于根据指定数据类型选取适当的函数来处理输入数据。
                sep：                <string> 默认的分隔符
                                            默认使用 \t


            工作流程：
                                               __init__(mode="w") ──► stage=0 ──► metadata_begin()
                                                                                         │
                                                                                         ▼
                   write_metadata(key, value) or writer.metadata=dict(key=value) ◄───► stage=1
                                                                                         │
                                                                                         ▼
                                                 contents_begin() ◄── stage=2 ◄── metadata_end()
                                                        │
                                                        ▼
                              __init__(mode="a") ──► stage=3 ◄───► write_contents(value) or writer.contents=value
                                                        │
                                                        ▼
                                                  contents_end() ───► stage=4

        """

        # 默认参数
        paras = {
            # 必要参数
            "file_path": None,
            # 写入模式相关参数
            "mode": "w",
            "paras_for_open": dict(encoding='utf-8'),
            # 可选参数
            "converter": CONVERTER_FOR_WRITER,
            "sep": "\t",
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        #
        paras["file_path"] = os.path.abspath(paras["file_path"])
        os.makedirs(os.path.split(paras["file_path"])[0], exist_ok=True)
        #
        assert paras["mode"] in ["w", "a"]
        #
        assert isinstance(paras["paras_for_open"], (dict,))
        paras["paras_for_open"]["mode"] = paras["mode"]
        #
        assert isinstance(paras["converter"], (Converter, dict,))
        assert isinstance(paras["sep"], (str,))

        self.paras = paras

        # metadata
        self.metadata = dict()
        # 文件对象
        self.file = None

        if paras["mode"] == "a":
            # 采用追加写模式
            # 尝试打开已有文件
            assert os.path.isfile(self.paras["file_path"])
            try:
                # 要求 metadata 已被写入
                reader = kevin_notation.Reader(file_path=self.paras["file_path"])
            except Exception as e:
                raise Exception(
                    f'file {self.paras["file_path"]} existed, but is not a standard kevin_toolbox document!')
            self.metadata = copy.deepcopy(reader.metadata)
            del reader
            # 获取文件对象
            self.file = open(self.paras["file_path"], **self.paras["paras_for_open"])
            # 初始状态码
            beg_stage = 3
        else:
            # 采用覆盖写模式
            self.metadata["sep"] = self.paras["sep"]
            # 获取文件对象
            self.file = open(self.paras["file_path"], **self.paras["paras_for_open"])
            # 写入文件标记
            self.file.write(f"# --kevin_notation--\n\n")
            # 初始状态码
            beg_stage = 0

        # 变量
        # state 状态码
        #   0：关闭所有写操作
        #   1：正在写 metadata
        #   2：正在写 contents
        self.state = dict(stage=beg_stage)

    # ------------------------------------ metadata ------------------------------------ #

    def metadata_begin(self):
        # 流程检查
        assert self.state["stage"] == 0, \
            f'Error: need to invoke __init__(mode="w") first!'
        self.state["stage"] = 1

        self.file.write(f"# --metadata--\n")
        self._write_metadata("sep", self.paras["sep"])
        self.file.flush()

    def _write_metadata(self, key, value):
        """
            通过键值对的形式写入一个 metadata

            参数：
                key：            键
                value：          值/值以及该键值对的其他信息。
                                    支持两种方式指定:
                                        <list or tuple> 直接指定 value 的值，写入方式参考默认值
                                        <dict> 一个包含 value 以及额外指定写入方式参数的字典
                                            {"value": <list or tuple>, "sep": ..., }
        """
        paras = dict()
        if isinstance(value, (dict,)):
            assert "value" in value
            paras = value
            value = paras.pop("value")
        elif not isinstance(value, (list, tuple,)):
            value = [value]
        assert isinstance(value, (list, tuple,))
        #
        sep = paras.get("sep", self.paras['sep'])
        assert isinstance(sep, (str,))

        # check
        if "column_num" in self.metadata:
            if key in ["column_name", "column_type"]:
                assert len(value) == self.metadata["column_num"]

        # write
        # key
        self.file.write(f"# {key}")
        if len(paras) > 0:
            self.file.write(f" ({','.join([f'{k}={v}' for k, v in paras.items()])})")
        self.file.write(f"\n")
        # value
        self.file.write(f"{sep.join([f'{v}' for v in value])}\n")

        # metadata
        value = list(value)
        self.metadata[key] = value[0] if len(value) == 1 else value
        #
        if key in ["column_name", "column_type"]:
            self.metadata["column_num"] = len(value)
        self.file.flush()

    def write_metadata(self, **kwargs):
        """
            写入 metadata（供外部调用）

            支持两种方式进行写入：
                write_metadata(key, value)
                write_metadata(metadata)    其中 metadata 是包含多组 key,value 的 dict。等效于多次调用第一种方式进行写入。

            注意，由于对默认分隔符的设定和写入已经在一开始的初始化和 metadata_begin() 时完成了，因此本函数将直接跳过/忽略 key="sep" 的情况。
        """
        # 流程检查
        if self.state["stage"] == 0:
            self.metadata_begin()
        assert self.state["stage"] == 1, \
            Exception(f"Error: please call metadata_begin() before _write_metadata!")

        paras = kwargs

        if "metadata" in paras:
            assert isinstance(paras["metadata"], (dict,))
            for key, value in paras["metadata"].items():
                if key != "sep":
                    self._write_metadata(key=key, value=value)
        elif "key" in paras:
            assert isinstance(paras["key"], (str,)) and "value" in paras
            if paras["key"] != "sep":
                self._write_metadata(key=paras["key"], value=paras["value"])
        else:
            raise ValueError("Requires parameter 'key, value' or 'metadata'")

    def metadata_end(self):
        # 流程检查
        assert self.state["stage"] == 1
        self.state["stage"] = 2

        self.file.write(f"\n")
        self.file.flush()

    # ------------------------------------ contents ------------------------------------ #

    def contents_begin(self):
        # 流程检查
        assert self.state["stage"] == 2, \
            f"Error: need to close the last metadata writing!"
        self.state["stage"] = 3

        self.file.write(f"# --contents--\n")
        self.file.flush()

    def _write_contents(self, row_ls):
        """
            通过 row_ls 以行的形式写入 contents

            参数：
                row_ls：          <list / list of list/tuple>  一行、或者多行的内容，将按照内部保存的 column_type 进行转换。
        """
        if len(row_ls) == 0:
            return

        row_ls = np.array(row_ls)
        if row_ls.ndim <= 1:
            row_ls = row_ls.reshape((1, -1))

        if "column_num" in self.metadata:
            assert row_ls.shape[-1] == self.metadata["column_num"], \
                f"{row_ls.shape}"

        # 转换并写入
        if "column_type" in self.metadata:
            type_ls = self.metadata["column_type"]
        else:
            type_ls = ["default"] * len(row_ls.shape[-1])

        for row in row_ls:
            row = [self.paras["converter"][type_](r) for type_, r in zip(type_ls, row)]
            line = f"{self.metadata['sep'].join(row)}\n"
            self.file.write(line)
        self.file.flush()

    def write_contents(self, **kwargs):
        """
            写入 contents（供外部调用）

            支持两种方式进行写入：
                write_contents(row_ls)         <list / list of list/tuple>  一行、或者多行的内容，将按照内部保存的 column_type 进行转换。
                write_contents(column_dict)    <dict of key:column_name value:list> 按照内部保存的 column_name 进行读取和写入。
        """
        # 流程检查
        if self.state["stage"] == 1:
            self.metadata_end()
        if self.state["stage"] == 2:
            self.contents_begin()
        assert self.state["stage"] == 3, \
            Exception(f"Error: please call contents_begin() before write_contents!")

        paras = kwargs

        if "row_ls" in paras:
            self._write_contents(row_ls=paras["row_ls"])
        elif "column_dict" in paras:
            assert isinstance(paras["column_dict"], (dict,))
            try:
                # 解释为多行
                row_ls = list(zip(*[paras["column_dict"][k] for k in self.metadata["column_name"]]))
            except:
                # 解释为单行
                row_ls = [[paras["column_dict"][k] for k in self.metadata["column_name"]], ]
            self._write_contents(row_ls=row_ls)
        else:
            raise ValueError("Requires parameter 'row_ls' or 'column_dict'")

    def contents_end(self):
        # 流程检查
        assert self.state["stage"] == 3
        self.state["stage"] = 4

        self.file.flush()

    # ------------------------------------ magic func ------------------------------------ #

    # self.key = value
    def __setattr__(self, key, value):
        """
            支持直接通过 self.key = value 的方式来写入 metadata 和 contents
        """
        if "state" not in self.__dict__:
            # state 未被设置，未完成初始化
            super().__setattr__(key, value)
        else:
            if self.state["stage"] == 1:
                if key == "metadata":
                    self.write_metadata(metadata=value)
                else:
                    self.write_metadata(key=key, value=value)
            elif self.state["stage"] == 3:
                if key == "row_ls":
                    self.write_contents(row_ls=value)
                elif key == "column_dict":
                    self.write_contents(column_dict=value)
                else:  # 兼容旧版本
                    warnings.warn(
                        f"Writing to contents with keys named other than row_ls and column_dict "
                        f"will no longer be supported in a future release", DeprecationWarning)
                    try:
                        self.write_contents(row_ls=value)
                    except:
                        self.write_contents(column_dict=value)
            else:
                raise ValueError(f"Error: please call metadata_begin() or contents_begin() before write!")

    # with 上下文管理器
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self

    def __del__(self):
        try:
            del self.paras, self.state
            self.file.close()
        except Exception as e:
            print(e)



import pytest
import numpy as np
from kevin_toolbox.patches.for_test import check_consistency
from kevin_toolbox.computer_science.algorithm.combinatorial_optimization import zero_one_knapsack_problem
from kevin_toolbox.computer_science.algorithm.for_seq import get_subsets


def test_zero_one_knapsack_problem():
    print("test combinatorial_optimization.zero_one_knapsack_problem()")

    for _ in range(1000):
        # 随机构建输入的 weights、upper_bound 和 values
        beg = int(np.random.randint(-10, 10))
        end = int(np.random.randint(1, 10))
        if end < beg:
            beg, end = end, beg
        weights = np.random.random(np.random.randint(1, 10)) * (end - beg) + beg
        values = np.random.random(len(weights)) * (end - beg) + beg
        # upper_bound = np.random.random() * end * 2.3
        upper_bound = (np.random.random() * (end - beg) + beg) * 2.3
        if np.random.randint(0, 2):
            weights = weights.astype(int)
            upper_bound = int(upper_bound)
            values = values.astype(int)

        # 使用 zero_one_knapsack_problem() 求解
        best_v_0, idx_ls_0 = zero_one_knapsack_problem(
            weights=weights.tolist(), values=values.tolist(), upper_bound=upper_bound,
        )

        # 通过暴力搜索获取真实答案
        best_v_1, idx_ls_s = None, []
        for w_ls, v_ls, idx_ls in zip(get_subsets(inputs=weights),
                                      get_subsets(inputs=values),
                                      get_subsets(inputs=range(len(weights)))):

            if sum(w_ls) > upper_bound:
                continue
            if best_v_1 is None or sum(v_ls) > best_v_1:
                best_v_1 = sum(v_ls)
                idx_ls_s = [idx_ls]
            elif sum(v_ls) == best_v_1:
                idx_ls_s.append(idx_ls)
        if best_v_1 is None:
            idx_ls_s = [None]

        # 检验
        # print(weights, values, upper_bound)
        # print(best_v_1, best_v_0)
        check_consistency(best_v_1, best_v_0)
        assert idx_ls_0 in idx_ls_s



import numpy as np


def cal_iou_between_boxes(**kwargs):
    """
        计算 boxes_0 和 boxes_1 之间的交并比 iou

        参数：
            boxes_0:          <3 axis np.array>
                                shape [batch_size, 2, dimensions]，各个维度的意义为：
                                    batch_size： 有多少个 box
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
            boxes_1:          <3 axis np.array>
                                shape 与 boxes_0 相同。
                                特别地，当 boxes_0 与 boxes_1 中某个的 batch_size 大于 1 时，另外一个可以将 batch_size 设为 1，
                                    此时将进行自动扩增。
            return_details: <boolean> 是否以详细信息的形式返回结果
                                默认为 False，此时返回：
                                    iou <np.array>
                                        shape [batch_size]
                                当设置为 True，将返回一个 dict：
                                    details = dict(
                                        iou=<np.array>,
                                        intersection=dict(areas=<np.array>, boxes=<np.array>,),  # areas 表示面积
                                        union=dict(areas=<np.array>),
                                        boxes_0=dict(areas=<np.array>, boxes=<np.array>,),
                                        boxes_1=dict(areas=<np.array>, boxes=<np.array>,),
                                    )
    """
    # 默认参数
    paras = {
        # 必要参数
        "boxes_0": None,
        "boxes_1": None,
        "return_details": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    for key in ("boxes_0", "boxes_1"):
        paras[key] = np.asarray(paras[key])
        assert paras[key].ndim == 3 and paras[key].shape[1] == 2
    assert paras["boxes_0"].shape[-1] == paras["boxes_1"].shape[-1]
    boxes_0, boxes_1 = paras["boxes_0"], paras["boxes_1"]
    boxes_0.sort(axis=1)
    boxes_1.sort(axis=1)

    # intersection
    beg = np.maximum(boxes_0[:, 0], boxes_1[:, 0])
    end = np.minimum(boxes_0[:, 1], boxes_1[:, 1])
    edge_lens = np.maximum(end - beg, 0)
    intersection = edge_lens.prod(axis=-1)

    # union
    areas_0 = (boxes_0[:, 1] - boxes_0[:, 0]).prod(axis=-1)
    areas_1 = (boxes_1[:, 1] - boxes_1[:, 0]).prod(axis=-1)
    union = areas_0 + areas_1
    union -= intersection

    # iou
    iou = intersection / np.maximum(union, 1e-10)

    if paras["return_details"]:
        details = dict(
            iou=iou,
            intersection=dict(areas=intersection, boxes=np.concatenate([beg[:, None, ...], end[:, None, ...]], axis=1)),
            union=dict(areas=union),
            boxes_0=dict(areas=areas_0, boxes=boxes_0),
            boxes_1=dict(areas=areas_1, boxes=boxes_1),
        )
        return details
    else:
        return iou



import numpy as np
from kevin_toolbox.math.dimension import coordinates, transpose


def unflatten_along_pattern(**kwargs):
    """
        将 x 最后的一个维度，按照 shape 对应的 generate_func 指定的遍历顺序进行堆叠
            实际上就是打平展开 flatten_along_pattern() 的逆向操作

        参数：
            x：                  <nparray/tensor>
            shape：              <list/tuple of integers>
            generate_func：      用于指定对 block 的遍历顺序
                                    默认使用 coordinates.generate(pattern="z_pattern", output_format="index_ls") 进行遍历
                                    你也可以自定义一个根据参数 shape 生成 zip_indices 格式的坐标列表的函数，来指定遍历顺序
    """
    # 默认参数
    paras = {
        # 必要参数
        "x": None,
        "shape": None,
        "generate_func": lambda shape: coordinates.generate(pattern="z_pattern", output_format="index_ls",
                                                            shape=shape),
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert paras["x"] is not None
    x = paras["x"]
    assert isinstance(paras["shape"], (list, tuple,)) and np.prod(paras["shape"]) == x.shape[-1], \
        f'type {type(paras["shape"])} in [list, tuple] ? {np.prod(paras["shape"])} == {x.shape[-1]}？'
    shape = paras["shape"]
    # generate_func
    assert callable(paras["generate_func"])

    # 获取转置的逆
    index_ls = paras["generate_func"](shape=shape)
    r_index_ls = transpose.get_inverse_index_ls(index_ls)

    # 对最后的维度进行转置
    y = transpose.inside_axis(x=x, axis=-1, index_ls=r_index_ls)

    # reshape
    new_shape = list(y.shape)[:-1] + list(shape)
    y = y.reshape(new_shape)

    return y



import pytest
from kevin_toolbox.patches.for_test import check_consistency

import os
import numpy as np

from kevin_toolbox.data_flow.file import kevin_notation
from kevin_toolbox.data_flow.file.kevin_notation.test.test_data.data_all import metadata_ls, content_ls, file_path_ls


@pytest.mark.parametrize("expected_metadata, expected_content, file_path",
                         zip(metadata_ls, content_ls, file_path_ls))
def test_reader(expected_metadata, expected_content, file_path):
    print("test Reader")

    # 读取
    chunk_size = np.random.randint(1, 10)
    with kevin_notation.Reader(file_path=file_path, chunk_size=chunk_size) as reader:
        # metadata
        # print(reader.metadata)
        check_consistency(expected_metadata, reader.metadata)
        # content
        content = next(reader)
        for chunk in reader:
            for key in content.keys():
                content[key].extend(chunk[key])
        # print(content)
        check_consistency(expected_content, content)


@pytest.mark.parametrize("expected_metadata, expected_content, file_path",
                         zip(metadata_ls, content_ls, file_path_ls))
def test_writer_0(expected_metadata, expected_content, file_path):
    """
        测试以 writer.key = value 的方式写入
            注意需要以在写入前后按照顺序依次调用 writer.metadata_begin()/end() 和 writer.contents_begin()/end()
    """
    print("test Writer")

    # 新建
    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data/temp",
                             os.path.basename(file_path))
    part = np.random.randint(1, 5)
    values = list(zip(*[expected_content[key][:-part] for key in expected_metadata["column_name"]]))
    with kevin_notation.Writer(file_path=file_path, mode="w", sep=expected_metadata["sep"]) as writer:
        writer.metadata_begin()
        if part % 2 == 0:
            # 逐个写入
            for key, value in expected_metadata.items():
                if key == "sep":
                    pass
                elif key == "column_name":
                    writer.column_name = value
                elif key == "column_type":
                    # 尝试使用局部指定的 sep
                    writer.column_type = {"value": value, "sep": " "}
                else:
                    writer.write_metadata(key=key, value=value)
        else:
            # 整体写入
            writer.metadata = expected_metadata
        writer.metadata_end()

        writer.contents_begin()
        writer.row_ls = values  # 列表方式写入
        writer.contents_end()

    # 续写
    # values = list(zip(*[expected_content[key][-part:] for key in expected_metadata["column_name"]]))
    values = {k: v[-part:] for k, v in expected_content.items()}
    with kevin_notation.Writer(file_path=file_path, mode="a") as writer:
        writer.column_dict = values  # 字典方式写入
        writer.contents_end()

    # 检验
    with kevin_notation.Reader(file_path=file_path, chunk_size=1000) as reader:
        # metadata
        check_consistency(expected_metadata, reader.metadata)
        # content
        content = next(reader)
        check_consistency(expected_content, content)


@pytest.mark.parametrize("expected_metadata, expected_content, file_path",
                         zip(metadata_ls, content_ls, file_path_ls))
def test_writer_1(expected_metadata, expected_content, file_path):
    """
        测试以 writer.write_metadata() 的方式写入
            可以省略 writer.metadata_begin()/end() 和 writer.contents_begin()/end()
    """
    print("test Writer")

    # 新建
    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data/temp",
                             os.path.basename(file_path))
    part = np.random.randint(1, 5)
    values = list(zip(*[expected_content[key][:-part] for key in expected_metadata["column_name"]]))
    with kevin_notation.Writer(file_path=file_path, mode="w", sep=expected_metadata["sep"]) as writer:
        if part % 2 == 0:
            # 逐个写入
            for key, value in expected_metadata.items():
                writer.write_metadata(key=key, value=value)
        else:
            # 整体写入
            writer.write_metadata(metadata=expected_metadata)

        writer.write_contents(row_ls=values)  # 列表方式写入

    # 续写
    values = {k: v[-part:] for k, v in expected_content.items()}
    with kevin_notation.Writer(file_path=file_path, mode="a") as writer:
        writer.write_contents(column_dict=values)  # 字典方式写入

    # 检验
    with kevin_notation.Reader(file_path=file_path, chunk_size=1000) as reader:
        # metadata
        check_consistency(expected_metadata, reader.metadata)
        # content
        content = next(reader)
        check_consistency(expected_content, content)


@pytest.mark.parametrize("expected_metadata, expected_content, file_path",
                         zip(metadata_ls, content_ls, file_path_ls))
def test_read(expected_metadata, expected_content, file_path):
    print("test read()")
    # 读取
    metadata, content = kevin_notation.read(file_path=file_path)
    # 检验
    check_consistency(expected_metadata, metadata)
    check_consistency(expected_content, content)


@pytest.mark.parametrize("expected_metadata, expected_content, file_path",
                         zip(metadata_ls, content_ls, file_path_ls))
def test_write(expected_metadata, expected_content, file_path):
    print("test write()")

    # 新建
    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data/temp", os.path.basename(file_path))

    # 列表方式写入
    kevin_notation.write(metadata=expected_metadata,
                         content=list(zip(*[expected_content[key] for key in expected_metadata["column_name"]])),
                         file_path=file_path)
    # 检验
    metadata, content = kevin_notation.read(file_path=file_path)
    check_consistency(expected_metadata, metadata)
    check_consistency(expected_content, content)

    # 字典方式写入
    kevin_notation.write(metadata=expected_metadata, content=expected_content, file_path=file_path)
    # 检验
    metadata, content = kevin_notation.read(file_path=file_path)
    check_consistency(expected_metadata, metadata)
    check_consistency(expected_content, content)



import numpy as np
from kevin_toolbox.math import utils


class Node:
    """
        计算图节点
            使用该节点类实例构建的计算图，将支持把节点中的张量 var 按照对应的 box 指示的位置（var在源张量中的位置），
            以行优先的内存顺序，进行拼接、展平和其逆向的分割。

        节点包含以下属性：
            sub_nodes:          <dict of node> 记录子节点。
                                        定义了计算图中节点之间的连接关系。
            var:                <np.array/torch.tensor> 节点值。
                                        用于保存节点上的计算结果。
            details:            <dict of paras> 节点属性。
                                        用于保存节点内的处理函数的参数，定义了节点的行为。
                                        包含以下参数：
                                            box_ls:     表示节点中的 var 是根据这些 box 从源张量中截取出来并拼接而成的。
                                                        each element is an array with shape [2, dimensions]
                                                        各个维度的意义为：
                                                            2：          box 的两个轴对称点
                                                            dimensions： 坐标的维度
                                            axis:       表示节点将根据 box_ls 中第 axis 个轴对 box_ls 和 var 进行分割，然后将分割结果
                                                            分别传递给对应的子节点。
                                            slice:      本节点的 var 和 box_ls 是如何从父节点中分割得到的。
                                            shape:      本节点中 var 的目标形状。
                （看到这里一脸懵逼很正常，不要担心，下面将结合具体的工作流程来介绍节点中各个属性的具体作用）

        使用方法：
            本节介绍如何使用本节点类，构建一个树状的计算图，并进行推理计算。

            1. 构建计算图：
                root_node = Node(box_ls=...)
                root_node.build_tree()  # 将从根节点出发，根据 box_ls 构建计算图

                假设输入的 box_ls = [ [[0, 0],
                                     [4, 4]],
                                    [[0, 4],
                                     [4, 16]],
                                    [[4, 0],
                                     [16, 4]],
                                    [[4, 4],
                                     [16, 16]] ] 则具体的构建过程为：
                首先从根节点出发：

                          root_node

                    box_ls=[box_0,1,2,3]
                    slice(None,None)
                    aixs=0

                其中 axis=0 表示我们先看 box 中的第0个轴（对应上面的第一列），可以看到该轴有两个分段，分别是 slice(0,4) 和 slice(4,16)。
                根据分段构建第一层子节点，将属于 slice(0,4) 的两个 box 分给 sub_node_0，依次类推：

                          root_node  ───────┬──────── sub_node_0
                                            │
                    box_ls=[box_0,1,2,3]    │      box_ls=[box_0,1]
                    slice(None,None)        │      slice(0,4)
                    aixs=0                  │
                                            │
                                            │
                                            │
                                            │
                                            │
                                            └──────── sub_node_1

                                                   box_ls=[box_2,3]
                                                   slice(4,16)

                对于第一层子节点，axis=1，看第1个轴，依次类推，构建第二层节点：
                                                                      ┌────── leaf_node_0
                                                                      │
                          root_node  ───────┬──────── sub_node_0 ─────┤    box_ls=[box_0]
                                            │                         │    slice(0,4)
                    box_ls=[box_0,1,2,3]    │      box_ls=[box_0,1]   │    axis=2
                    slice(None,None)        │      slice(0,4)         │
                    aixs=0                  │      axis=1             └────── leaf_node_1
                                            │
                                            │                              box_ls=[box_1]
                                            │                              slice(4,16)
                                            │                              axis=2
                                            │
                                            └──────── sub_node_1 ─────┬────── leaf_node_2
                                                                      │
                                                   box_ls=[box_2,3]   │    box_ls=[box_2]
                                                   slice(4,16)        │    slice(0,4)
                                                   axis=1             │    axis=2
                                                                      │
                                                                      └────── leaf_node_3

                                                                           box_ls=[box_3]
                                                                           slice(4,16)
                                                                           axis=2

                当 axis==box.shape[-1] 时，到达叶节点，停止构建。此时叶节点中应该有且只有一个 box。

            2. 检查并初始化计算图：
                root_node.init_tree()  # 从叶节点出发，检查各个节点的合法性，并初始化相关的变量，依次向上直至根节点。

                检查叶节点中 box 数量为1，并计算 box 的 size，依次向上累加：

                                                                      ┌────── leaf_node_0
                                                                      │
                          root_node  ◄──────┬──────── sub_node_0 ◄────┤    box_ls=[box_0]
                                            │                         │    slice(0,4)
                    box_ls=[box_0,1,2,3]    │      box_ls=[box_0,1]   │    axis=2
                    slice(None,None)        │      slice(0,4)         │    size=4*4=16
                    aixs=0                  │      axis=1             └────── leaf_node_1
                    size=64+192=256         │      size=16+48=64
                                            │                              box_ls=[box_1]
                                            │                              slice(4,16)
                                            │                              axis=2
                                            │                              size=4*12=48
                                            └──────── sub_node_1 ◄────┬────── leaf_node_2
                                                                      │
                                                   box_ls=[box_2,3]   │    box_ls=[box_2]
                                                   slice(4,16)        │    slice(0,4)
                                                   axis=1             │    axis=2
                                                   size=48+144=192    │    size=12*4=48
                                                                      └────── leaf_node_3

                                                                           box_ls=[box_3]
                                                                           slice(4,16)
                                                                           axis=2
                                                                           size=12*12=144

            3. 使用计算图进行推理：
                本计算图支持两种计算

                split(beg_axis)  # 从根节点出发，根据节点属性 details 中的参数，对节点值 var 进行分割，然后将分割的片段分别传递给后继的子节点中的 var。
                                依次向下迭代，直至叶节点。
                                最后分割的结果保存在叶节点的 var 中，可以通过 get_leaf_nodes() 获取子节点并读取其值。

                以上面的计算图为例，下面展示一次典型的 split() 计算过程：
                假设根节点中 var.shape=[10,256,3]，从源张量的第1个轴开始分割，亦即 beg_axis=1。

                首先，根节点读取其子节点的 size，根据 root_node->size / sub_node->size 的比例，确定 root_node.var 在 beg_axis + aixs 对应的维度上的分割比例。
                然后进行分割。

                          root_node  ───────┬───────► sub_node_0
                    var:[10,256,3]          │      var=root_node.var[10,:64,3]
                    box_ls=[box_0,1,2,3]    │      box_ls=[box_0,1]
                    slice(None,None)        │      slice(0,4)
                    aixs=0                  │      axis=1
                    size=256                │      size=64
                                            │
                                            │
                                            │
                                            │
                                            └───────► sub_node_1
                                                   var=root_node.var[10,64:256,3]
                                                   box_ls=[box_2,3]
                                                   slice(4,16)
                                                   axis=1
                                                   size=192

                再将子节点的 var 的第 beg_axis + aixs 个维度，reshape 成 [length of sub_node.slice,-1] 。
                （如果最后的 -1 维度计算结果为 1，则将该维度去除）

                          root_node  ───────┬───────► sub_node_0
                    var:[10,256,3]          │      var:[10,4,16,3]
                    box_ls=[box_0,1,2,3]    │      box_ls=[box_0,1]
                    slice(None,None)        │      slice(0,4)
                    aixs=0                  │      axis=1
                    size=256                │      size=64
                                            │
                                            │
                                            │
                                            │
                                            └───────► sub_node_1
                                                   var:[10,12,16,3]
                                                   box_ls=[box_2,3]
                                                   slice(4,16)
                                                   axis=1
                                                   size=192

                依次类推，得到：

                                                                      ┌─────► leaf_node_0
                                                                      │    var:[10,4,4,3]
                          root_node  ───────┬───────► sub_node_0 ─────┤    box_ls=[box_0]
                    var:[10,256,3]          │      var:[10,4,16,3]    │    slice(0,4)
                    box_ls=[box_0,1,2,3]    │      box_ls=[box_0,1]   │    axis=2
                    slice(None,None)        │      slice(0,4)         │
                    aixs=0                  │      axis=1             └─────► leaf_node_1
                                            │                              var:[10,4,12,3]
                                            │                              box_ls=[box_1]
                                            │                              slice(4,16)
                                            │                              axis=2
                                            │
                                            └───────► sub_node_1 ─────┬─────► leaf_node_2
                                                   var:[10,12,16,3]   │    var:[10,12,4,3]
                                                   box_ls=[box_2,3]   │    box_ls=[box_2]
                                                   slice(4,16)        │    slice(0,4)
                                                   axis=1             │    axis=2
                                                                      │
                                                                      └─────► leaf_node_3
                                                                           var:[10,12,12,3]
                                                                           box_ls=[box_3]
                                                                           slice(4,16)
                                                                           axis=2



                concat(beg_axis)  # 从叶节点出发，将叶节点中的 var 向上传递给父节点。父节点根据 details 对从子节点中获取的 var_ls 进行合并，
                                    保存到本节点的 var 中，然后继续将本节点的 var 传递给更上层的节点。依次向上迭代，直至根节点。
                                    最后总的合并结果保存在根节点的 var 中。

                split 和 concat 互为逆操作。

                本类还提供了针对 split 和 concat 操作的进一步封装：
                    10
                    20
                这两个函数相较于原始的 split() 等函数提供了更完善的接口检查，建议使用这两个函数。
    """

    def __init__(self, **kwargs):
        """
            设定关键参数

            参数：
                box_ls:         <list of np.arrays>
                                    each element is an array with shape [2, dimensions]
                                    各个维度的意义为：
                                        2：          box 的两个轴对称点
                                        dimensions： 坐标的维度
                                    要求：
                                        - 各个 box 应该是已经 sorted 的，亦即小坐标在前大坐标在后。
                                            例如 box=[[1,2],[0,4]] 是错误的。
                                            而 box=[[0,2],[1,4]] 是合法的。
                                        - 各个 box 在坐标轴上的投影之间没有重叠部分
                                        函数 geometry.for_boxes.boolean_algebra() 返回的 boxes 结果，以及
                                        函数 geometry.for_boxes.detect_collision() 返回的 node_ls 中每个 node 下面
                                        的 node.description["by_boxes"] 都符合该要求。
                    （为节省计算量，本函数不对 box_ls 的合法性进行完整的检查。）
        """

        # 默认参数
        paras = {
            # 必要参数
            "box_ls": [],
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert isinstance(paras["box_ls"], (list,))
        if len(paras["box_ls"]) > 0:
            assert paras["box_ls"][0].ndim == 2 and paras["box_ls"][0].shape[1] == 2

        self.sub_nodes = dict()  # {<node>.slice[0]: <node>, ...}  分割起始点到子节点的映射
        self.var = None
        self.details = dict(
            #
            box_ls=paras["box_ls"],
            # 潜变量（根据 box_ls 推断得到，不需要外部设定）
            axis=0,  # 本节点根据该轴进行分割。本节点的子节点（若有）将根据该轴的下一个轴进行分割，因此本参数随着树的深度增加而增加，该参数也暗含了树的深度信息。
            slice=[None, None],
            size=None,
        )

    def build_tree(self):
        """
            构建计算图
        """
        assert len(self.details["box_ls"]) > 0, \
            f'Error: box_ls of node with depth {self.details["axis"]} is empty! ' \
            f'please invoke `node.details["box_ls"]=box_ls` before build_tree!'

        # 到达叶节点。（递归出口）
        if self.details["axis"] >= self.details["box_ls"][0].shape[-1]:
            return

        # 添加子节点。
        for box in self.details["box_ls"]:
            new_slice = box[:, self.details["axis"]].tolist()
            if new_slice[0] not in self.sub_nodes:
                node = Node()
                node.details["slice"] = new_slice
                node.details["axis"] = self.details["axis"] + 1
                self.sub_nodes[new_slice[0]] = node
            node = self.sub_nodes[new_slice[0]]
            assert new_slice == node.details["slice"], \
                f'{new_slice} == {node.details["slice"]} ？\n' \
                f'Error: Please check whether the projection of the input box_ls on each coordinate axis overlaps'
            node.details["box_ls"].append(box)

        for node in self.sub_nodes.values():
            node.build_tree()

    def init_tree(self):
        """
            检查并初始化计算图
        """
        # 先初始化子节点
        for node in self.sub_nodes.values():
            node.init_tree()

        # 定义每个节点的初始化行为
        if len(self.sub_nodes) == 0:
            # 叶节点
            #   检查叶节点的合法性
            assert len(self.details["box_ls"]) == 1
            #   根据 box 计算 size
            box = self.details["box_ls"][0]
            self.details["size"] = np.prod(box[1] - box[0])
        else:
            # 非叶节点
            #   累加子节点的 size
            self.details["size"] = sum([node.details["size"] for node in self.sub_nodes.values()])

    def print_tree(self):
        """
            依照DFS打印树
        """
        tap = "\t" * 2 * self.details["axis"]
        print(f'{tap}depth:{self.details["axis"]} {self.details}')
        print(f'{tap}------------ divided by axis {self.details["axis"]} ------------')
        for node in self.sub_nodes.values():
            node.print_tree()

    def clear_tree(self, depths):
        """
            清理节点的 var

            参数:
                depths:     <list/tuple/set of integers>  要清除哪些深度的层
        """
        assert isinstance(depths, (list, tuple, set,))

        depths = set(depths)
        if self.details["axis"] in depths:
            depths.remove(self.details["axis"])
            self.var = None

        if len(depths) > 0:
            for node in self.sub_nodes.values():
                node.clear_tree(depths)

    def get_leaf_nodes(self):
        """
            以列表形式返回所有叶节点
        """
        leaf_nodes = []
        if len(self.sub_nodes) == 0:
            leaf_nodes.append(self)
        else:
            for node in self.sub_nodes.values():
                leaf_nodes.extend(node.get_leaf_nodes())
        return leaf_nodes

    def split(self, beg_axis=0):
        """
            按照 BFS 的顺序，逐层向下分割节点中的 var

            参数：
                beg_axis:       <integer> 从 var 的第几个轴开始分割。
        """
        assert isinstance(beg_axis, (int,))
        assert self.details["size"] is not None, \
            f"Error: computational graph not initialized! please invoke init_tree() before inference!"
        assert self.var is not None, \
            f'before splitting node with depth {self.details["axis"]}, its var needs to be assigned a value'

        axis = beg_axis + self.details["axis"]
        slices = [slice(None, None)] * (axis + 1)
        beg = 0
        for _, node in sorted(self.sub_nodes.items(), key=lambda x: x[0]):
            # 截取出属于各个子节点的部分 var
            if node.details["size"] * self.details["size"] == 0:
                end = beg
            else:
                end = beg + int(round(node.details["size"] / self.details["size"] * self.var.shape[axis]))
            slices[axis] = slice(beg, end)
            #
            sub_var = self.var[tuple(slices)]
            # 对截取的部分按照子节点的要求进行 reshape
            sub_shape = list(self.var.shape)
            sub_span = node.details["slice"][1] - node.details["slice"][0]
            sub_shape[axis] = sub_span
            if len(node.sub_nodes) > 0:
                # 对于非叶节点，需要再增加一个维度留给其子节点进行分割
                if sub_span > 0:  # 考虑 sub_span=0 的情况
                    assert node.details["size"] % sub_span == 0
                    sub_shape.insert(axis + 1, node.details["size"] // sub_span)
                else:
                    sub_shape.insert(axis + 1, 0)
            #
            node.var = sub_var.reshape(sub_shape)
            #
            beg = end

        for node in self.sub_nodes.values():
            node.split(beg_axis=beg_axis)

    def concat(self, beg_axis=0):
        """
            按照 DFS 的顺序，从叶节点逐层向上合并节点中的 var

            参数：
                beg_axis:       <integer> 叶节点中的 box 指定的坐标是从 var 的第几个 axis 开始对应的
                                例如： beg_axis=1 时，box=[[i,j],[m,n]] 表示叶节点中的 var 是从原张量的 [:, i:m, j:n, ...] 部分截取出来的。
        """

        if len(self.sub_nodes) == 0:
            # 叶节点
            box = self.details["box_ls"][0]
            assert self.var is not None \
                   and list(self.var.shape[beg_axis:beg_axis + box.shape[-1]]) == (box[1] - box[0]).tolist()
        else:
            # 非叶节点
            #   让子节点进行合并
            for node in self.sub_nodes.values():
                node.concat(beg_axis=beg_axis)
            #   再合并子节点中的 var
            var_ls = [node.var for _, node in sorted(self.sub_nodes.items(), key=lambda x: x[0])]
            end_axis = beg_axis + self.details["axis"]
            _, function_table = utils.get_function_table_for_array_and_tensor(var_ls[0])
            concat, flatten = function_table["concat"], function_table["flatten"]
            # 合并
            res = concat(var_ls, axis=end_axis)
            # 打平
            if beg_axis < end_axis:
                # 将已 concat 的部分进行展平，展平后的 ndim 将减1
                res = flatten(res, axis_0=end_axis - 1, axis_1=end_axis)
            self.var = res



import pytest
from kevin_toolbox.env_info.version import compare, parse_to_array, sort_ls


@pytest.mark.parametrize("v_0, operator, v_1, kwargs, result",
                         [([1, 2, 3], ">=", [1, 3], dict(), False),
                          ([1, 2, 3], "<", [1, 3], dict(), True),
                          ([1, 2, 3], "==", [1, 2, 3, 0, 0], dict(), True),
                          ("1.10.0a0", ">", "1.2", dict(), True),
                          ("1.10.0a0", "<=", "1.2", dict(), False),
                          # 测试 sep
                          ("1.2.3", "<", [1, 3], dict(sep='.'), True),
                          ("n_7.3", "<", [1, 3], dict(sep='.'), True),
                          ("1_2_3", "==", "1_2_3_0_0", dict(sep='_'), True),
                          # 测试 mode
                          ("1_2_3", "==", "1_2_3_4_5", dict(sep='_', mode='short'), True), ])
def test_compare(v_0, operator, v_1, kwargs, result):
    assert compare(v_0, operator, v_1, **kwargs) == result


@pytest.mark.parametrize("string, sep, result",
                         [("1,2,3", ',', [1, 2, 3]),
                          ("1_2.3", '.', [-1, 3]),
                          ("neg", '_', [-1])])
def test_parse_to_array(string, sep, result):
    assert parse_to_array(string, sep=sep) == result


@pytest.mark.parametrize("inputs, expected",
                         [(dict(version_ls=["0.10.7", (0, 7), [0, 7, 5]], reverse=False),
                           [(0, 7), [0, 7, 5], "0.10.7"]),
                          (dict(version_ls=["0.10.7", (0, 7), [0, 7, 5]], reverse=True),
                           ["0.10.7", [0, 7, 5], (0, 7)]), ])
def test_sort_ls(inputs, expected):
    assert sort_ls(**inputs) == expected



import numpy as np
from .generate_shuffled_index_ls import generate_shuffled_index_ls
from .generate_z_pattern_indices_ls import generate_z_pattern_indices_ls
from ... import coordinates

SUPPORTED_PATTERNS = {"z_pattern", "shuffle_inside_block", "normal"}
SUPPORTED_FORMATS = {"index_ls", "indices_ls", "zip_indices"}


def generate_integrated_api(**kwargs):
    """
        整合了已有生成方法的集成接口
            按照不同模式 pattern 对 shape 进行遍历，并生成指定格式的 坐标列表

        参数：
            shape:              <list/tuple of integers> 坐标所属的多维变量的形状。
            pattern:            <str> 生成/遍历坐标的模式
                                    目前支持：
                                        "normal":
                                            生成 从原点出发按照行/列（在kwargs中通过order可以指定）优先遍历的下标列表
                                            核心调用的是 normal_indices_generator()
                                        "z_pattern" :
                                            生成 从原点出发进行之字形（Z形）遍历 的下标列表。
                                            核心调用的是 generate_z_pattern_indices_ls()
                                        "shuffle_inside_block" :
                                            对 shape 内的各个 block 中的坐标进行打乱，生成随机打乱的下标列表。
                                            核心调用的是 generate_shuffled_index_ls()
            kwargs:             <dict> 其他补充参数。
            output_format:      <str> 输出的目标格式。
                                    目前支持的坐标格式：
                                        "index_ls" , "indices_ls" , "zip_indices"
                                        各种格式的具体定义参见 coordinates.convert()
    """

    # 默认参数
    paras = {
        # 必要参数
        "shape": None,
        "pattern": None,
        "output_format": None,
        #
        "kwargs": dict(),
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert isinstance(paras["shape"], (list, tuple,)) and len(paras["shape"]) > 0
    #
    assert paras["pattern"] in SUPPORTED_PATTERNS, \
        f'Unknown pattern!\ncurrently supported patterns: {SUPPORTED_PATTERNS}'
    #
    assert paras["output_format"] in SUPPORTED_FORMATS, \
        f'Unknown format!\ncurrently supported formats: {SUPPORTED_FORMATS}'
    #
    assert isinstance(paras["kwargs"], (dict,))

    # 生成坐标
    if paras["pattern"] == "z_pattern":
        var = generate_z_pattern_indices_ls(shape=paras["shape"], **paras["kwargs"])
        input_format = "indices_ls"
    elif paras["pattern"] == "shuffle_inside_block":
        var = generate_shuffled_index_ls(shape=paras["shape"], **paras["kwargs"])
        input_format = "index_ls"
    else:  # "normal"
        var = np.asarray(list(coordinates.normal_indices_generator(shape=paras["shape"], **paras["kwargs"])))
        input_format = "indices_ls"

    # 转换格式
    res = coordinates.convert(var=var,
                              input_format=input_format, output_format=paras["output_format"],
                              shape=paras["shape"])

    return res



import os
import shutil


def remove(path, ignore_errors=False):
    """
        移除文件/文件夹

        返回:
            boolean 是否成功
    """
    try:
        if os.path.isfile(path):  # 移除文件
            os.remove(path)
        elif os.path.isdir(path):  # 移除文件夹
            shutil.rmtree(path=path, ignore_errors=False)
        else:
            raise FileNotFoundError(f'path: {path} not exists')
        return True
    except Exception as e:  # 删除失败
        if not ignore_errors:
            raise Exception(e)
        return False


if __name__ == '__main__':
    remove(path="233.txt", ignore_errors=True)



import os
import subprocess
import argparse
import version

"""
检查当前版本
- 若在可用版本中，有比当前版本更高的版本，则更新到可以获取到的最新版本。
"""

out_parser = argparse.ArgumentParser(description='check_version_and_update')
out_parser.add_argument('--package_name', type=str, required=True)
out_parser.add_argument('--cur_version', type=str, required=False)
out_parser.add_argument('--available_versions', nargs='+', type=str, required=False)
out_parser.add_argument('--verbose', type=int, required=False, default=1)
args = out_parser.parse_args().__dict__

# try to read cur_version
if args["cur_version"] is None:
    ex = subprocess.Popen(f'pip list | grep "{args["package_name"]} "', shell=True, stdout=subprocess.PIPE)
    out, _ = ex.communicate()
    out = out.decode().strip()
    # breakpoint()
    args["cur_version"] = out.split(args["package_name"])[-1].strip()

# try to read available versions
if args["available_versions"] is None:
    ex = subprocess.Popen(f'pip install {args["package_name"]}==?', shell=True, stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT)
    out, _ = ex.communicate()
    out = out.decode().strip()
    if "(from versions:" in out:
        v_ls = out.split("(from versions:")[-1].rsplit(")", 1)[0].split(",", -1)
        v_ls = [i.strip() for i in v_ls]
    else:
        v_ls = ["none"]
    args["available_versions"] = version.sort_ls(version_ls=v_ls, reverse=True)

if len(args["available_versions"]) > 0 and version.compare(args["available_versions"][0], ">", args["cur_version"]):
    ex = subprocess.Popen(
        f'pip install {args["package_name"]}=={args["available_versions"][0]} --no-dependencies',
        shell=True, stdout=subprocess.PIPE
    )
    out, _ = ex.communicate()
    res = out.decode().strip()
else:
    res = "Already the latest version, no need to update"

if args["verbose"]:
    # print(args)
    print(res)



from abc import ABC, abstractmethod


class Sequence_Map_Base(ABC):

    @abstractmethod
    def __getitem__(self, item):
        return self[item]

    @abstractmethod
    def __len__(self):
        return len(self)



from .flatten_along_pattern import flatten_along_pattern as flatten
from .unflatten_along_pattern import unflatten_along_pattern as unflatten
from .split_blocks import split_blocks
from .merge_blocks import merge_blocks



import numpy as np
from .parse_version import parse_version_string_to_array as parse_to_array


def compare_version(v_0, operator, v_1, **kwargs):
    """
        在两个版本号之间比较大小

        参数：
            v_0, v_1:               <array of integers / string of version>
            operator：               比较符
                                        支持的取值： "==", ">=", "<=", ">", "<"
        可选参数：
            mode：                   对齐方式
                                        "long"（默认）： 以 v_0, v_1 中较长的为基准进行对齐，对于较短的版本号，缺省部分将补 0
                                        "short"： 以较短的为基准进行对齐，对于较长的版本号，多余部分直接截断
                                        （建议选用 "long"，因为对于无法解释的部分 version，parse_to_array()
                                        默认返回 [..., -1, ...]，此时如果使用 "short" 可能引发意外的错误）
            sep：                    分隔符
                                        对于string类型的版本号，默认使用"."作为分隔符，但是也可以在 kwargs 中通过 sep 指定，
                                        具体参考 parse_to_array()
        返回：
            boolean 比较结果
    """
    assert operator in {"==", ">=", "<=", ">", "<"}
    v_0, v_1 = map(lambda x: parse_to_array(x, **kwargs) if isinstance(x, (str,)) else x, [v_0, v_1])
    v_0, v_1 = map(lambda x: np.array(x, dtype=int).reshape(-1), [v_0, v_1])
    mode = kwargs.get("mode", "long")
    assert mode in ["long", "short"]

    if mode == "long":
        gap = np.zeros(shape=max(len(v_0), len(v_1)))
        gap[:len(v_0)] += v_0
        gap[:len(v_1)] -= v_1
    else:
        gap = np.zeros(shape=min(len(v_0), len(v_1)))
        gap += v_0[:len(gap)]
        gap -= v_1[:len(gap)]
    first_larger_index = np.argmax(gap > 0) if np.any(gap > 0) else len(gap)
    first_smaller_index = np.argmax(gap < 0) if np.any(gap < 0) else len(gap)
    if first_larger_index == first_smaller_index:
        res = "="
    elif first_larger_index < first_smaller_index:
        res = ">"
    else:
        res = "<"

    return res in operator


if __name__ == '__main__':
    print(compare_version([1, 2, 3], ">=", [1, 3]))
    print(compare_version("1.10.0a0", ">", "1.2"))
    print(compare_version("1.10.0a0", "<=", "1.2"))



import os
from functools import wraps


def restore_original_work_path(func):
    """
        装饰器，在运行函数 func 前备份当前工作目录，并在函数运行结束后还原到原始工作目录。
    """
    @wraps(func)
    def wrap(*args, **kwargs):
        pwd_bak = os.getcwd()
        res = func(*args, **kwargs)
        os.chdir(pwd_bak)
        return res

    return wrap



from kevin_toolbox.geometry import for_boxes
import copy
import numpy as np


class Node:
    def __init__(self):
        # 用矢量描述
        self.description = dict(
            by_item_ids=dict(
                difference=set(),
                intersection=set(),
            ),
            by_boxes=None,
        )


def detect_overlap(**kwargs):
    """
        在将 boxes_ls 中的每个 boxes 视为一个 item，检测所有 item 之间的重叠区域

        参数：
            boxes_ls:       <3 axis np.array> 需要检测的 item
                                each boxes inside the list is an np.array with shape [batch_size, 2, dimensions]
                                各个维度的意义为：
                                    batch_size： 有多少个 box
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
        输出：
            node_ls:        <list of Node> 由于重叠分割出的不同区域
                                每个区域由一个 node 表示，其中：
                                    node.description["by_item_ids"]["intersection"]     <set of item_id> 这个区域是由哪些 item 相交而成的
                                    node.description["by_item_ids"]["difference"]       <set of item_id> 在上面交集的基础上，应该减去哪些 item
                                    node.description["by_boxes"]                        <boxes> 该区域由哪些 box 组成
                                可见 node.description["by_item_ids"] 描述了该区域的“来源”，
                                node.description["by_boxes"] 描述了该区域的形状。
                注意：
                    - 将排除所有体积为0的区域
                    - node_ls 中除了 item 之间的重叠区域，也记录了各个 item 的独有区域
    """
    # 默认参数
    paras = {
        # 必要参数
        "boxes_ls": None,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert isinstance(paras["boxes_ls"], (list,))
    boxes_ls = []
    for boxes in paras["boxes_ls"]:
        if boxes is None or for_boxes.cal_area(boxes=boxes) == 0:
            boxes_ls.append(None)
        else:
            boxes_ls.append(boxes)

    # 计算碰撞情况
    collision_groups = for_boxes.detect_collision(boxes_ls=boxes_ls, complexity_correction_factor_for_aixes_check=1.0,
                                                  duplicate_records=True)
    # 按照潜在的包含关系进行排序
    # 目标：保证 collision_groups 中后一个 id 指向的 item 不可能被之前的包含
    # 条件：
    #   - （充分条件）如果两个 item 之间具有包含关系，比如 item_0 included by item_1，那么必然有 item_0 相关的碰撞 pairs 数量少于 item_1 的。
    #       如果将配装关系看做 graph，则应该按 节点度数 进行排序，最大度数在前。
    #   - （必要条件）比较两个 item 的外切 box 的体积，体积大的在前。
    #       （这个条件相较于上一个更加严格，但是由于该条件需要计算面积，代价较大，因此可以先使用第一个条件进行初筛以减少计算量，
    #       但是为什么我不这样做呢？因为我懒，完毕。）
    #   - （改进的必要条件）上面的条件有可能出现体积为0的情况，尤其是对于高维的 box 而言。此时无法保证排序。
    #       因此改为对外切 box 的每条边的边长进行比较。
    sorted_collisions = [(k, v_ls) for k, v_ls in collision_groups.items()]
    sorted_collisions.sort(
        key=lambda x: tuple(np.max(boxes_ls[x[0]][:, 1, :], axis=0) - np.min(boxes_ls[x[0]][:, 0, :], axis=0)),
        reverse=True)

    # 为所有 item 创建 node
    item_node_ls = []
    input_node_ls = [None] * len(boxes_ls)
    output_node_ls = []
    ids = set(collision_groups.keys())
    for i, boxes in enumerate(boxes_ls):
        node = Node()
        node.description["by_item_ids"]["intersection"].add(i)
        node.description["by_boxes"] = boxes
        item_node_ls.append(node)
        if i in ids:
            # 有碰撞，需要计算
            input_node_ls[i] = node
        else:
            # 完全没有碰撞，可以直接输出
            if boxes is not None:
                output_node_ls.append(node)

    recursion(collisions=sorted_collisions, input_node_ls=input_node_ls, output_node_ls=output_node_ls)

    return output_node_ls


def recursion(collisions, input_node_ls, output_node_ls):
    if len(collisions) == 0:
        return
    k, v_ls = collisions[0]

    # 1. 处理 items[k]
    node = copy.deepcopy(input_node_ls[k])
    #
    boxes_ls = [node.description["by_boxes"]]
    boxes_ls.extend([input_node_ls[v].description["by_boxes"] for v in v_ls])
    binary_operation_ls = ["diff"] * (len(boxes_ls) - 1)
    res_boxes = for_boxes.boolean_algebra(boxes_ls=boxes_ls, binary_operation_ls=binary_operation_ls)
    #
    node.description["by_item_ids"]["difference"].update(v_ls)
    node.description["by_boxes"] = res_boxes

    if res_boxes is not None:
        output_node_ls.append(node)

    # 2. 处理 items[k] 内
    # items[v] intersect with items[k]
    # 构建 node_ls
    inner_node_ls = [None] * len(input_node_ls)
    included_box_ids = set()  # 用于记录被 items[k] 包含了的 items[v]，被包含的 items[v] 将不参与到后续的“处理 items[k] 外”流程
    for v in v_ls:
        res_boxes = for_boxes.boolean_algebra(
            boxes_ls=[input_node_ls[k].description["by_boxes"], input_node_ls[v].description["by_boxes"]],
            binary_operation_ls=["and"]
        )
        #
        if for_boxes.boolean_algebra(
                boxes_ls=[input_node_ls[v].description["by_boxes"], res_boxes],
                binary_operation_ls=["diff"]
        ) is None:
            # items[k] 包含了 items[v]
            # 第 v 个 item 发生的碰撞全部都在第 k 个 item 内
            # 因此后续不需要计算位于 items[k] 以外的 items[v] 的残余部分与其他 items 的碰撞情况。
            # 注意：此处不能通过 交集的体积 是否等于 被减数的体积 来判断包含关系，因为可能存在某个维度边长大小为0，而导致所有体积都为0的情况。
            included_box_ids.add(v)
        # intersection
        node = copy.deepcopy(input_node_ls[v])
        # 相交部分仅需考量两个节点都有的 difference
        node.description["by_item_ids"]["difference"].intersection_update(
            input_node_ls[k].description["by_item_ids"]["difference"])
        #
        node.description["by_item_ids"]["intersection"].add(k)
        node.description["by_boxes"] = res_boxes
        inner_node_ls[v] = node
    # 构建 collisions
    inner_collisions = []
    for i, j_ls in collisions:
        if i in v_ls:
            inner_collisions.append((i, j_ls.difference({k}).intersection(v_ls)))
    # 递归处理
    recursion(collisions=inner_collisions, input_node_ls=inner_node_ls, output_node_ls=output_node_ls)

    # 3. 处理 items[k] 外
    # items[v] reduced by items[k]
    # 构建 node_ls 和 collisions
    outer_node_ls = [None] * len(input_node_ls)
    outer_collisions = []
    need_to_be_excluded = included_box_ids.union({k})
    for i, j_ls in collisions:
        if i in need_to_be_excluded:
            continue
        outer_node_ls[i] = input_node_ls[i]
        outer_collisions.append((i, j_ls.difference(need_to_be_excluded)))
    for v in v_ls.difference(need_to_be_excluded):
        res_boxes = for_boxes.boolean_algebra(
            boxes_ls=[outer_node_ls[v].description["by_boxes"], input_node_ls[k].description["by_boxes"]],
            binary_operation_ls=["diff"]
        )
        #
        outer_node_ls[v].description["by_item_ids"]["difference"].add(k)
        outer_node_ls[v].description["by_boxes"] = res_boxes
    # 递归处理
    recursion(collisions=outer_collisions, input_node_ls=outer_node_ls, output_node_ls=output_node_ls)


if __name__ == '__main__':
    from test.test_data.data_1 import boxes

    # boxes = boxes[[0, 1, 2, 3], ...]
    print(boxes.shape)

    output_node_ls = detect_overlap(boxes_ls=[box.reshape(1, 2, box.shape[-1]) for box in boxes])

    counter = 0
    print(len(output_node_ls))
    for node in output_node_ls:
        print(node.description)
        if node.description["by_boxes"] is not None:
            counter += len(node.description["by_boxes"])
    print(counter)



import math
import numpy as np
from kevin_toolbox.math.dimension import coordinates

DEFAULT_RD = np.random.RandomState()


def generate_shuffled_index_ls(**kwargs):
    """
        对 shape 内的各个 block 中的坐标进行打乱，生成随机打乱的 index_ls
            index_ls 的具体定义参见 coordinates.convert()
            参照卷积的形式，按照 kernel_size 和 stride 对 shape 的最后几个维度进行遍历，依次对遍历经过的 block 进行内部随机打乱。
            支持通过指定 generate_indices_ls_func 来设置对 block 的遍历顺序/模式。

        参数：
            shape:                      形状
            kernel_size：                卷积核的大小
                                            默认为 None 表示使用整个 shape 作为 kernel_size
            stride：                     卷积的步长
                                            默认为 None 表示与 kernel_size 相等
            allow_duplicates:           允许出现重复值。
                                            默认为 False
            generate_func_of_traversal_order_for_blocks：  用于指定对 block 的遍历顺序
                                            默认使用 coordinates.generate(pattern="z_pattern", output_format="indices_ls") 进行遍历
                                            你也可以自定义一个根据参数 shape 生成 zip_indices 格式的坐标列表的函数，来指定遍历顺序
            seed:                       随机种子
            rd:                         随机生成器
                                            （默认不需要设置）

        blocks 相关变量的计算公式：
            第i个维度上 block 的数量 n_i = argmin_{n_i} stride[i] * n_i + kernel[i] >= shape[i]  s.t. n_i>=0
                            等效于： n_i = math.ceil( max( shape[i]-kernel[i], 0 ) / stride[i] )
            对于由 (n_i,n_j,n_k) 指定的 crop，它在原始feather map上的坐标为：
                                crop = x[ stride[i] * n_i: stride[i] * n_i + kernel[i], ... ]
    """
    # 默认参数
    paras = {
        # 必要参数
        "shape": None,
        #
        "kernel_size": None,
        "stride": None,
        "allow_duplicates": False,
        #
        "generate_func_of_traversal_order_for_blocks":
            lambda shape: coordinates.generate(pattern="z_pattern", output_format="indices_ls", shape=shape),
        #
        "seed": None,
        "rd": DEFAULT_RD,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    # shape
    if isinstance(paras["shape"], (int,)):
        paras["shape"] = [paras["shape"]]
    assert isinstance(paras["shape"], (tuple, list,)) and 0 < len(paras["shape"])
    shape = list(paras["shape"])
    # kernel_size
    if paras["kernel_size"] is None:
        paras["kernel_size"] = shape
    if isinstance(paras["kernel_size"], (int,)):
        paras["kernel_size"] = [paras["kernel_size"]] * len(shape)
    assert isinstance(paras["kernel_size"], (tuple, list,)) and 0 < len(paras["kernel_size"]) <= len(shape)
    kernel_size = list(paras["kernel_size"])
    # stride_ls
    if paras["stride"] is None:
        paras["stride"] = kernel_size
    if isinstance(paras["stride"], (int,)):
        paras["stride"] = [paras["stride"]] * len(kernel_size)
    assert isinstance(paras["stride"], (tuple, list,)) and len(paras["stride"]) == len(kernel_size)
    stride_ls = list(paras["stride"])
    # generate_func_of_traversal_order_for_blocks
    assert callable(paras["generate_func_of_traversal_order_for_blocks"])
    # 随机生成器
    assert paras["rd"] is not None
    rd = paras["rd"]
    rd.seed(paras["seed"])

    # 构建索引矩阵
    index_ls = np.arange(0, np.prod(shape))
    index_ls = index_ls.reshape(shape)
    if len(kernel_size) < len(shape):
        new_shape = [-1] + shape[-len(kernel_size):]
        index_ls = index_ls.reshape(new_shape)
    else:
        index_ls = index_ls[np.newaxis, :]

    # 构建遍历顺序
    shape_of_blocks = []
    for w, k, s in zip(shape[-len(kernel_size):], kernel_size, stride_ls):
        shape_of_blocks.append(math.ceil(max(w - k, 0) / s) + 1)
    indices = paras["generate_func_of_traversal_order_for_blocks"](shape=shape_of_blocks)
    beg_indices = indices * np.array([stride_ls])
    end_indices = beg_indices + np.array([kernel_size])

    #
    if paras["allow_duplicates"]:
        shuffle_func = lambda u: rd.choice(u.reshape(-1), u.size,
                                           replace=paras["allow_duplicates"]).reshape(u.shape)
    else:
        # permutation 稍微快一点
        shuffle_func = lambda u: rd.permutation(u.reshape(-1)).reshape(u.shape)

    # 随机打乱
    for i in range(index_ls.shape[0]):
        for beg, end in zip(beg_indices, end_indices):
            x = index_ls[i]
            slices = tuple([slice(int(b), int(e)) for b, e in zip(beg, end)])
            crop = x[slices]
            crop[:] = shuffle_func(crop[:])
            #
    index_ls = index_ls.reshape(-1)

    return index_ls


if __name__ == '__main__':
    shape = [6, 6]
    index_ls = generate_shuffled_index_ls(shape=shape, stride=[3, 3], kernel_size=[3, 3], seed=114)
    print(index_ls.reshape(shape))

    # from line_profiler import LineProfiler
    # import time
    #
    # lp = LineProfiler()
    #
    # lp_wrapper = lp(generate_shuffled_index_ls)
    # lp_wrapper(shape=[8181], seed=1145141919)
    # lp.print_stats()
    #
    # time_start = time.time()
    # generate_shuffled_index_ls(shape=[818100], seed=1145141919)
    # time_end = time.time()
    # print('time cost', time_end - time_start, 's')



from .kevin_notation_writer import Kevin_Notation_Writer as Writer
from .kevin_notation_reader import Kevin_Notation_Reader as Reader
from .converter import Converter, CONVERTER_FOR_READER, CONVERTER_FOR_WRITER
from .read import read
from .write import write



from mpl_toolkits.mplot3d import Axes3D  # 3d坐标系
import numpy as np
from .arrow3d import Arrow3D


def add_trajectory_3d(ax3d, points, draw_key_points=True, scale=None):
    """
        参数：
            ax3d:
            points:                 <np.array> 坐标点 shape [batch, 3]
            draw_key_points:        <boolean> 是否绘制关键点
            scale:                  <float> 放大系数
                                        不设定时，默认根据 points 推断得到的最小分辨率来设置。
    """
    assert isinstance(ax3d, (Axes3D,))
    points = np.asarray(points)
    assert len(points) >= 1 and points.ndim == 2 and points.shape[-1] == 3, \
        f'expected shape of points is [batch, 3], but got a {points.shape}'

    if scale is None:
        if len(points) > 1:
            temp = points.copy()
            temp[1:] -= temp[:-1]
            temp[0] -= points[-1]
            scale = np.min(np.sum(temp ** 2, axis=1)) ** 0.5
        else:
            scale = 1

    # 箭头
    for beg, end in zip(points[:-1], points[1:]):
        ax3d.add_artist(Arrow3D(*beg.tolist(), *(end - beg).tolist(),
                                arrowstyle="-|>", mutation_scale=int(15 * scale),
                                color="blue", alpha=0.5))
        # ax3d.quiver(*beg.tolist(), *(end - beg).tolist(), length=1, normalize=True)

    # 关键点
    if draw_key_points:
        ax3d.scatter3D(xs=points[:, 0], ys=points[:, 1], zs=points[:, 2],
                       marker='o', s=scale * 25,  # 散点的形状及大小
                       c="red", alpha=0.9,  # 散点的颜色及透明度
                       cmap=None)

    return ax3d



import torch
from .convert import convert_to_numpy
from kevin_toolbox.patches.for_torch.compatible import where as torch_where

# 计算设备（尽量使用gpu来加速计算）
device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')


def merge_cfm_ls(cfm_ls, to_numpy=True, **kwargs):
    """
        将多个混淆矩阵进行合并

        参数：
            cfm_ls:                 list of confusion_matrices，每个 confusion matrices 都是一个包含
                                    thresholds、tp_ls、tn_ls、fp_ls、fn_ls 等字段的 dict
                                            其具体介绍参见函数 binary_classification.cal_cfm()
            to_numpy（可选）:           boolean，是否将结果从 tensor 转换为 numpy
                                            默认为True
        工作流程：
            我们以合并两个 confusion_matrices 为例，
                对于递增的 tp_ls 字段，
                    首先计算它的每两个相邻的 threshold 之间 tp 的差值：
                        diff_tp_ls = tp_ls.copy()
                        diff_tp_ls[1:] -= tp_ls[:-1]  # 其中 diff_tp_ls[m] 表示 threshold=thresholds[m] 时刚好取到的 tp 样本
                    然后根据 thresholds，对两个 diff_tp_ls 进行合并，比如：
                        thresholds_0：   [0.9, 0.8, 0,7]
                        diff_tp_ls_0：   [1, 4, 5]
                        thresholds_1：   [0.8, 0.6, 0,5]
                        diff_tp_ls_1：   [2, 3, 7]
                    那么合并后就是：
                        thresholds：     [0.9, 0.8, 0,7, 0.6, 0,5]
                        diff_tp_ls：     [1, 4+2, 5, 3, 7]
                    最后再根据 diff_tp_ls，使用 cumsum 累积出合并后的 tp_ls：
                        tp_ls:          [1, 7, 12, 15, 22]
                对于递增的 fp_ls 字段也有类似的操作，
                对于递减的 tn_ls 和 fn_ls 字段，则根据 tp_ls、fp_ls 得出。
    """
    assert isinstance(cfm_ls, (list,)) and len(cfm_ls) > 0

    # 合并
    res = dict()
    temp = dict()
    for key in {"thresholds", "tp_ls", "fp_ls"}:
        tensor_ls = []
        for prediction in cfm_ls:
            if isinstance(prediction[key], torch.Tensor):
                tensor_ls.append(prediction[key].to(device))
            else:
                tensor_ls.append(torch.tensor(prediction[key], device=device))
        res[key] = torch.cat(tensor_ls, dim=0)
        temp[key] = tensor_ls

    # 计算 diff
    for key in {"tp_ls", "fp_ls"}:
        beg = 0
        for tensor_ in temp[key]:
            end = beg + tensor_.shape[0]
            res[key][beg + 1:end] -= tensor_[: - 1]
            beg = end

    # 按照 thresholds 从大到小进行排序，然后进行累加
    res["thresholds"], sorted_indices = torch.sort(res["thresholds"], descending=True, dim=0)
    for key in {"tp_ls", "fp_ls"}:
        res[key] = torch.cumsum(res[key][sorted_indices.reshape(-1)], dim=0, dtype=torch.int64)

    # 去除相同 threshold 下的结果
    diff_thresholds = res["thresholds"].clone()
    diff_thresholds[:-1] -= res["thresholds"][1:]  # 取梯度
    diff_thresholds[-1] = -1  # 保证最后一个元素被下面的 torch.where 取出
    diff_indices = torch_where(diff_thresholds != 0)
    for key, value in res.items():
        res[key] = value[diff_indices]

    # 计算 tn_ls 和 fn_ls
    res["tn_ls"] = res["fp_ls"][-1] - res["fp_ls"]
    res["fn_ls"] = res["tp_ls"][-1] - res["tp_ls"]

    if to_numpy:
        res = convert_to_numpy(res)

    return res



def binary_search(ls, value, is_sorted=False):
    """
        二分法查找
            返回给定的 value 在已经排序好的数组 ls 中，按照顺序应该插入到哪个 index 位置上
            比如 ls=[0, 1, 2, 2, 3], value=2 则返回适合插入的第一个位置 index=2

        参数:
            ls:             <list/tuple>
            value:
            is_sorted:      <boolean> 数组是否已经按从小到大进行排序

        返回：
            index
    """
    assert isinstance(ls, (list, tuple,))
    if not is_sorted:
        ls = sorted(ls)
    return _binary_search(ls=ls, value=value, beg=0, end=len(ls) - 1)


def _binary_search(ls, value, beg, end):
    if beg > end:
        return beg
    mid = (beg + end) // 2
    if value <= ls[mid] and (mid - 1 < 0 or ls[mid - 1] < value):
        return mid
    elif value < ls[mid]:
        return _binary_search(ls, value, beg=beg, end=mid - 1)
    else:
        return _binary_search(ls, value, beg=mid + 1, end=end)



from kevin_toolbox.developing.temperate import My_Iterator
from kevin_toolbox.computer_science.data_structure import Executor


def build_iterator(executor_ls, size_ls=None):
    """
        根据 executor_ls 构建数据集的迭代器

        参数：
            executor_ls：        执行器列表
            size_ls：            数据集大小
                                    当为非空时，将根据 size_ls 对 executor 实际上生成数据集的大小进行检查
        返回：
            iterator：          一个产生 dataset 的迭代器
    """
    return Iterator_for_Executor_ls(executor_ls, size_ls)


class Iterator_for_Executor_ls(My_Iterator):
    """
        根据 executor_ls 构建迭代器
            支持的功能参见 My_Iterator_Base：
                支持 iter(self) 迭代
                    self.set_range(beg,end) 设定迭代起点、终点
                    self.pass_by(num)    跳过若干个值
                支持 self[index] 取值、len(self) 获取长度属性

        参数：
            executor_ls：        执行器列表
            size_ls：            数据集大小
                                    当为非空时，将根据 size_ls 对 executor 实际上生成数据集的大小进行检查
    """

    def __init__(self, executor_ls, size_ls=None):
        super().__init__(executor_ls)  # 在父类方法中， self.array = executor_ls

        if size_ls is not None:
            assert isinstance(size_ls, (list, tuple,)) and len(size_ls) == len(self.array)
        self.size_ls = size_ls

    def read(self, index):
        executor = self.array[index]
        assert isinstance(executor, (Executor,))

        res = executor()
        if self.size_ls is not None:
            assert len(res[list(res.keys())[0]]) == self.size_ls[index]
        return res



import torch
from kevin_toolbox.patches.for_torch.math import my_around
from kevin_toolbox.patches.for_torch.compatible import where as torch_where
from .convert import convert_to_numpy

# 计算设备（尽量使用gpu来加速计算）
device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')


def cal_cfm(scores, labels, to_numpy=True, decimals_of_scores=None, **kwargs):
    """
        将 scores 中所有不同的取值作为模型的 threshold，然后统计出在不同 threshold 下模型的预测结果中，
        TP、FN、FP、TN的数量。

        统计 TP、FN、FP、TN 的原理：
            首先依据 scores 从大到小对 labels 和 scores 进行排序，
            对于给定的一个阈值 threshold，我们选取 scores >= threshold 的部分对应的 labels，
            假设前 m 个满足阈值，
            则 labels[:m] 的部分会被模型判断为 positive，在这一部分的样本中
                label 为 0 是假阳性 FP 样本，
                label 为 1 是真阳性 TP 样本，
            则剩余的 labels[m:] 的部分会被模型判断为 negative，在这一部分的样本中
                label 为 0 是真阴性 TN 样本，
                label 为 1 是假阴性 FN 样本。

        参数：
            scores:             list of floats indicating the confidence of the model to predict the sample
                                as a positive sample.
                                表示模型将样本预测为正样本的置信度
                                    shape [sample_nums, 1]
            labels:             list of 0/1 describing the true label of the sample
                                描述样本的真实标签，其中1表示正样本，0表示负样本。
                                    shape [sample_nums, 1]
            to_numpy（可选）:     boolean，是否将结果从 tensor 转换为numpy
                                    默认为True
            decimals_of_scores（可选）：     integer，scores 要保留的小数位数（直接截取而非四舍五入）
                                    实际上，scores的取值有非常多，
                                    为了避免 thresholds 过多，可以通过限制 scores 的小数位，来对结果进行进一步合并和去重，
                                    从而减少内存占用。
                                    默认为 None，表示不对 scores 进行操作。
        返回：
            confusion_matrices： 描述了不同 threshold 取值下，二分类混淆矩阵中的各项 TP、FN、FP、TN 的取值，
                                是包含以下字段的一个 dict：
                                    thresholds：         scores 中不同的取值，从大到小排列
                                    tp_ls：              在对应的 threshold 取值下，有多少个样本是 true positive
                                    tn_ls：              true negative
                                    fp_ls：              false positive
                                    fn_ls：              false negative
                                    （以上的 shape 都是 [diff_nums]，其中 diff_nums 表示 scores 中有多少个不同的元素）
    """
    assert len(labels.shape) == len(scores.shape) == 2 and labels.shape[-1] == scores.shape[-1] == 1 and labels.shape[
        0] == scores.shape[0], \
        Exception(f"Error: The shapes of scores ({scores.shape}) and labels ({labels.shape}) are not equal, "
                  f"or do not satisfy the form of [sample_nums, 1]")

    scores_ = torch.tensor(scores, device=device, dtype=torch.float32) if not isinstance(scores,
                                                                                         torch.Tensor) else scores
    labels_ = torch.tensor(labels, device=device, dtype=torch.float32) if not isinstance(labels,
                                                                                         torch.Tensor) else labels
    if decimals_of_scores is not None:
        scores_ = my_around(scores_, decimals=decimals_of_scores, inplace=scores_ is not scores, floor=True)

    # 按照 scores 从大到小进行排序
    sorted_scores, sorted_indices = torch.sort(scores_, descending=True, dim=0)
    sorted_labels = labels_[sorted_indices.reshape(-1)]

    # 统计取前 m 个预测为正时，模型的预测结果
    # 其中 tp_ls[m] 表示前 m 个样本中，都多少个是label为真的。
    tp_ls = torch.cumsum(sorted_labels, dim=0, dtype=torch.int64)
    fp_ls = torch.cumsum(sorted_labels == 0, dim=0, dtype=torch.int64)

    # 找出 scores 不同的取值下的结果
    diff_scores = sorted_scores.clone()
    diff_scores[:-1] -= sorted_scores[1:]  # 取梯度
    diff_scores[-1, 0] = -1  # 保证最后一个元素被下面的 torch.where 取出
    diff_indices = torch_where(diff_scores != 0)
    #
    tp_ls = tp_ls[diff_indices]
    fp_ls = fp_ls[diff_indices]

    res = dict(thresholds=sorted_scores[diff_indices],
               tp_ls=tp_ls, fp_ls=fp_ls,
               tn_ls=fp_ls[-1] - fp_ls, fn_ls=tp_ls[-1] - tp_ls)

    if to_numpy:
        res = convert_to_numpy(res, decimals_of_scores)

    return res



import copy
import torch
import numpy as np
from kevin_toolbox.computer_science.algorithm.statistician._init_var import _init_var


class Average_Accumulator:
    """
        用于计算平均值的累积器
    """

    def __init__(self, **kwargs):
        """
            参数：
                data_format:            指定数据格式
                like:                   指定数据格式
                    指定输入数据的格式，有三种方式：
                        1. 显式指定数据的形状和所在设备等。
                            data_format:        <dict of paras>
                                    其中需要包含以下参数：
                                        type_:              <str>
                                                                "numpy":        np.ndarray
                                                                "torch":        torch.tensor
                                        shape:              <list of integers>
                                        device:             <torch.device>
                                        dtype:              <torch.dtype>
                        2. 根据输入的数据，来推断出形状、设备等。
                            like:               <torch.tensor / np.ndarray / int / float>
                        3. 均不指定 data_format 和 like，此时将等到第一次调用 add()/add_sequence() 时再根据输入来自动推断。
                        以上三种方式，默认选用最后一种。
                        如果三种方式同时被指定，则优先级与对应方式在上面的排名相同。
        """

        # 默认参数
        paras = {
            # 指定输入数据的形状、设备
            "data_format": None,
            "like": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        #
        self.paras = paras
        self.var = _init_var(like=paras["like"], data_format=paras["data_format"])
        self.state = dict(
            total_nums=0,
        )

    def add_sequence(self, var_ls):
        for var in var_ls:
            self.add(var)

    def add(self, var):
        """
            添加单个数据

            参数:
                var:                数据
        """
        if self.var is None:
            self.var = _init_var(like=var)
        # 累积
        self.var += var
        self.state["total_nums"] += 1

    def get(self):
        """
            获取当前累加的平均值
                当未有累积时，返回 None
        """
        if len(self) == 0:
            return None
        return self.var / len(self)

    def clear(self):
        self.var = _init_var(like=self.var)
        self.state = dict(
            total_nums=0,
        )

    def __len__(self):
        return self.state["total_nums"]


if __name__ == '__main__':
    seq = list(torch.tensor(range(1, 10)))
    avg = Average_Accumulator()
    for i, v in enumerate(seq):
        avg.add(var=v)
        print(i, v, avg.get())



from .get_function_table_for_array_and_tensor import get_function_table_for_array_and_tensor
from .convert_dtype import convert_dtype
from .get_crop_by_box import get_crop_by_box
from .set_crop_by_box import set_crop_by_box
from .spilt_integer_most_evenly import spilt_integer_most_evenly



from .convert import convert_format as convert
from .generate import generate_integrated_api as generate
from .generate import generate_shuffled_index_ls, generate_z_pattern_indices_ls
from .generate import normal_indices_generator



import hashlib
import json
import warnings


def get_hash(item, length=None, mode="md5"):
    """
        将输入 item 序列化，再计算出 hash 值
            本函数对于无序的字典能够提供唯一的 hash 值

        参数:
            item:
            length:         <int> 返回 hash 字符串的前多少位
                                默认为 None，返回所有结果
            mode:           <str> 支持 hashlib 中提供的 hash 方法
                                当设置为 None 时候表示不使用 hash 处理而直接返回序列化后的字符串
    """
    assert mode is None or mode in hashlib.__all__

    try:
        hash_ = json.dumps(item, sort_keys=True).encode('utf-8')
    except:
        warnings.warn(
            f"the item {type(item)} unable to be serialized by json, reproducibility is no longer guaranteed!",
            UserWarning
        )
        hash_ = f"{item}".encode('utf-8')
    if mode is not None:
        worker = eval(f'hashlib.{mode}()')
        worker.update(hash_)
        hash_ = worker.hexdigest()
    hash_ = hash_ if length is None else hash_[:length]
    return hash_


if __name__ == '__main__':
    print(get_hash(item={2, 4}, mode="sha512"))
    print(get_hash(item=[2, 4], mode="sha512"))



import threading
from collections import Iterable

'''
https://mp.weixin.qq.com/s?src=11&timestamp=1650039124&ver=3740&signature=Sm*QyFsWmuR2JoIrPqqsaOl6*r2pt2e53sHM2bVOjrm8ELxOfdvo9fRXHYAUZA-xwgHL4tDa*FjaQ9IGbEtPBKhPn4z553V0QdVuZ8ISXAfSMh-FBtCVwyhoKhdxlm6Z&new=1
https://www.cda.cn/view/123373.html
https://zhuanlan.zhihu.com/p/29379247
'''


class Producer:
    def __init__(self, **kwargs):
        # 默认参数
        paras = {
            # 必要参数
            "creator": None,
            # 仓库大小
            "lower_bound_of_storage_size": None,
            # 锁
            "lock": threading.Lock(),
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert isinstance(paras["creator"], (Iterable,))

        self.paras = paras
        self.storage = []

    def run(self):
        pass

    def get_lock(self):
        pass



from .confusion_matrices.cal_cfm import cal_cfm
from .confusion_matrices.merge_cfm_ls import merge_cfm_ls
from .confusion_matrices.cal_cfm_iteratively_by_chunk import cal_cfm_iteratively_by_chunk
from .confusion_matrices.cal_tpr_and_fpr import cal_tpr_and_fpr
from .confusion_matrices.accumulator_for_cfm import Accumulator_for_Confusion_Matrices as Accumulator_for_Cfm
from .confusion_matrices.convert import convert_to_numpy



import torch
import numpy as np
from kevin_toolbox.math.transform import dct
from kevin_toolbox.math import utils


class DCT_Calculator:
    """
        多维dct变换
            对张量的最后几个维度进行dct变换或者逆变换

        使用方法：
            calculator = dct.Calculator(...)  # 可预设使用的转换矩阵
            outputs = calculator(inputs, reverse, ...)
        更多请参考 calculator.cal() 函数的介绍

        ps：
            - 本模块计算DCT时并没有使用类似FFT的动态规划方式来节省计算量，因为本模块更多地关注使用gpu并行计算的场景，而
                诸如文章 https://jz.docin.com/p-699413364.html 中的快速DCT都难以实行并行计算。
                因而对于 basis_series_num 较小（能够被gpu一次性装下并计算）的情况，快速DCT的实际速度较慢。
                以后有可能会针对cpu的场景，增加快速DCT的计算方式。
            - 本模块支持 torch.tensor/np.array 类型的输入，并且会将输入变量所在的设备来作为计算设备。
                因此如果需要使用 gpu 进行计算，请首先保证输入变量已经指定到某个 gpu 设备上了。
    """

    def __init__(self, **kwargs):
        """
            参数：
                sampling_points_num_ls:     <list of integers> 对应维度上，进行转换时，采样点数量，的列表
                basis_series_num_ls:        <list of integers> 对应维度上，进行转换时，使用的基函数数量，的列表
                        以上参数均可在 cal() 时再具体指定，即使提前设定也不会提前生成相关转换矩阵
        """
        # 默认参数
        paras = {
            "sampling_points_num_ls": None,
            "basis_series_num_ls": None,
        }
        # 获取参数
        paras.update(kwargs)

        # 校验参数
        for k in ["sampling_points_num_ls", "basis_series_num_ls"]:
            assert paras[k] is None or isinstance(paras[k], (list, tuple,))

        self.paras = paras

        # 变量
        self.trans_matrix_repos = dict()  # {device: {sampling_points_num/r_num: trans_matrix, ...}, ...}

    def cal(self, **kwargs):
        """
            多维dct变换
                对张量的最后几个维度进行dct变换或者逆变换

            参数：
                x:                          <torch.tensor/np.array> 输入张量
                reverse:                    <boolean> 是否进行逆变换
                sampling_points_num_ls:     <list of integers> 对应维度上，进行转换时，采样点数量，的列表
                                                不设置时，默认使用初始化时设置的值，
                                                如果进一步连初始化时也没有设置时，将尝试根据 x 和 basis_series_num_ls 推断得到
                basis_series_num_ls:        <list of integers> 对应维度上，进行转换时，使用的基函数数量，的列表
                                                不设置时，默认使用初始化时设置的值，
                                                如果进一步连初始化时也没有设置时，将尝试根据 x 和 sampling_points_num_ls 推断得到

            例子：
                在 reverse=False 正向模式下时，当输入为 x [b, n0, n1, n2] 时，
                    在设置 sampling_points_num_ls=[n0, n1, n2] 和 basis_series_num_ls=[m0, m1, m2] 下，
                    将对输入的最后 len(basis_series_num_ls)=3 个维度进行变换，得到 y [b, m0, m1, m2]

            注意：
                - 基函数的数量 basis_series_num 不应超过采样点的数量 sampling_points_num
                - 当基函数的数量 basis_series_num 小于采样点的数量 sampling_points_num 时，此时转换过程是有损的，将丢失高频信息
                - 本函数将输入变量 x 所在的设备来作为计算设备。因此如果需要使用 gpu 进行计算，请首先保证输入变量已经指定到某个 gpu 设备上了。

            建议：
                - 对于 np.array 类型的输入和 dtype!=torch.float32 的 torch.tensor 类型的输入，
                    本函数会先转换成 <torch.tensor with dtype=float32> 再进行计算，
                    因此直接使用 <torch.tensor with dtype=float32> 类型输入可以跳过该转换过程，从而实现加速。

            返回：
                y：          <torch.tensor with dtype=float32> （所在设备与输入变量保持一致）
        """
        # 默认参数
        paras = {
            # 必要参数
            "x": None,
            #
            "reverse": False,
            "sampling_points_num_ls": self.paras["sampling_points_num_ls"],
            "basis_series_num_ls": self.paras["basis_series_num_ls"],
        }
        # 获取参数
        paras.update(kwargs)

        # 校验参数
        if not torch.is_tensor(paras["x"]):
            device = torch.device("cpu")
            x = torch.tensor(paras["x"], dtype=torch.float32, device=device)
        else:
            device = paras["x"].device
            x = paras["x"].to(dtype=torch.float32)
        #
        k0, k1 = ["sampling_points_num_ls", "basis_series_num_ls"] if paras["reverse"] else ["basis_series_num_ls",
                                                                                             "sampling_points_num_ls"]
        # 正向时，参数 basis_series_num_ls 是必要的
        # 反向时，参数 sampling_points_num_ls 是必要的
        assert paras[k0] is not None, \
            f"missing required parameter {k0}"
        # 正向时，sampling_points_num_ls 应该与 x 最后几个维度保持一致
        # 反向时，basis_series_num_ls 应该与 x 最后几个维度保持一致
        if paras[k1] is not None:
            assert list(paras[k1]) == list(x.shape[-len(paras[k1]):])
        else:
            paras[k1] = list(x.shape[-len(paras[k0]):])
        #
        for k in [k0, k1]:
            assert (np.array(paras[k]) > 0).all()
        assert len(paras[k0]) == len(paras[k1]) > 0
        # basis_series_num 不应超过采样点的数量 sampling_points_num
        assert (np.array(paras["sampling_points_num_ls"]) - np.array(paras["basis_series_num_ls"]) >= 0).all(), \
            f"basis_series_num should not exceed the number of sampling points sampling_points_num"
        if device not in self.trans_matrix_repos:
            self.trans_matrix_repos[device] = dict()

        "计算"
        for i, (r_num, c_num) in enumerate(
                zip(reversed(paras["sampling_points_num_ls"]), reversed(paras["basis_series_num_ls"]))):
            # 获取转换矩阵
            if r_num not in self.trans_matrix_repos[device]:
                # 新增到库
                self.trans_matrix_repos[device][r_num] = torch.tensor(dct.generate_trans_matrix(shape=[r_num, r_num]),
                                                                      dtype=torch.float32, device=device)
            B = self.trans_matrix_repos[device][r_num][..., :c_num]
            # 将待处理维度放到最后
            x = x.transpose(-1 - i, -1)
            # 变换
            if not paras["reverse"]:
                x = x @ B
            else:
                x = x @ B.t()
        # 还原维度顺序
        new_axis_idx = list(range(x.ndim))
        new_axis_idx.insert(-len(paras["sampling_points_num_ls"]), new_axis_idx[-1])
        new_axis_idx.pop(-1)
        y = x.permute(*new_axis_idx)

        return y

    def __call__(self, x, **kwargs):
        kwargs["x"] = x
        return self.cal(**kwargs)


if __name__ == '__main__':
    a = torch.ones([100, 200, 3, 4, 5], dtype=torch.float32)
    calculator = dct.Calculator()
    b = calculator(a, reverse=False, basis_series_num_ls=[3, 2, 4])
    a1 = calculator(b, reverse=True, sampling_points_num_ls=[3, 4, 5])
    print(b.shape)
    print(a1.shape)

    from line_profiler import LineProfiler

    lp = LineProfiler()
    lp_wrapper = lp(calculator.cal)
    lp_wrapper(x=a, reverse=False, basis_series_num_ls=[3, 2, 4])
    lp.print_stats()



import re


def get_value_by_name(var, name):
    """
        通过解释名字得到取值方式，然后到 var 中获取对应部分的值。

        参数：
            var:            任意支持索引取值的变量
            name:           <string> 取值方式。
                                由多组 "<解释方式><键>" 组成。
                                解释方式支持以下几种:
                                    "@"     使用 eval() 读取键
                                    ":"     使用 str() 读取键
                                    "|"     依次尝试 str() 和 eval() 两种方式
                                示例:
                                    "@100"      表示读取 var[eval("100")]
                                    ":epoch"    表示读取 var["epoch"]
                                    "|1+1"    表示首先尝试读取 var["1+1"]，若不成功则尝试读取 var[eval("1+1")]
                                假设 var=dict(acc=[0.66,0.78,0.99])，如果你想读取 var["acc"][1] => 0.78，那么可以将 name 写成：
                                    ":acc@1" 或者 "|acc|1" 等。
                                注意，在 name 的开头也可以添加任意非解释方式的字符，本函数将直接忽略它们，比如下面的:
                                    "var:acc@1" 和 "xxxx|acc|1" 也能正常读取。
    """
    assert isinstance(name, (str,))

    key_ls = re.split('[:@|]', name)
    # assert len(key_ls) > 0
    flag_idx = len(key_ls[0])
    for key in key_ls[1:]:
        flag = name[flag_idx]
        flag_idx += len(key) + 1
        #
        if flag == "@":
            var = var[eval(key)]
        elif flag == "|":
            try:
                var = var[key]
            except:
                var = var[eval(key)]
        else:  # ":"
            var = var[key]

    return var



import os
import json
import copy
from kevin_toolbox.data_flow.file.json_.converter import integrate
from kevin_toolbox.computer_science.algorithm.for_nested_dict_list import traverse


def write_json(content, file_path, sort_keys=False, converters=None):
    file_path = os.path.abspath(file_path)
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    if converters is not None:
        converter = integrate(converters)
        content = traverse(var=[copy.deepcopy(content)],
                           match_cond=lambda _, __, v: not isinstance(v, (list, dict)), action_mode="replace",
                           converter=lambda _, x: converter(x))[0]

    with open(file_path, 'w') as f:
        json.dump(content, f, indent=4, ensure_ascii=False, sort_keys=sort_keys)



import numpy as np
from .detect_collision_inside_boxes import detect_collision_inside_boxes


def detect_collision_among_boxes_ls(**kwargs):
    """
        将 boxes_ls 中的每个 boxes 视为一个 item，进行碰撞检测
            本函数是在 detect_collision_between_boxes() 的基础上实现的，具体工作原理请参见该函数。

        参数：
            boxes_ls:       <list of boxes/None> 需要检测的 item
                                each boxes inside the list is an np.array with shape [batch_size, 2, dimensions]
                                各个维度的意义为：
                                    batch_size： 有多少个 box
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
                                支持使用 None 作为占位符，标记为 None 的 item 将不与任一其他 item 发生碰撞
            complexity_correction_factor_for_aixes_check:   <float/integer>
                                    参见 detect_collision_between_boxes() 的介绍
            duplicate_records:  <boolean>
                                    参见 detect_collision_between_boxes() 的介绍
        输出：
            collision_groups:   <dict of integers set> 检出的碰撞对。
                                其中的第 i 个 set 记录了 id==i 的 boxes 表示的 item 与其他哪些 items 存在碰撞
                                    比如：  collision_groups[0] = {1, 2} 表示0号 item 与1、2号 items 发生了碰撞
    """
    # 默认参数
    paras = {
        # 必要参数
        "boxes_ls": None,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert isinstance(paras["boxes_ls"], (list,))
    boxes_ls = paras["boxes_ls"]
    boxes_ls_wto_none = [boxes for boxes in boxes_ls if boxes is not None]

    if len(boxes_ls_wto_none) <= 1:
        # 单个物体没有发生碰撞
        return dict()
    # 对所有的 box 进行碰撞检测，然后消除同一个 item 内部的碰撞
    item_id_ls = [np.ones(shape=boxes.shape[0], dtype=int) * i for i, boxes in enumerate(boxes_ls) if boxes is not None]
    boxes_com = np.concatenate(boxes_ls_wto_none, axis=0)
    item_id_com = np.concatenate(item_id_ls)  # 记录各个 box 所属的 item
    collision_groups_com = detect_collision_inside_boxes(boxes=boxes_com, **kwargs)
    #
    collision_groups = dict()
    for k, v_ls in collision_groups_com.items():
        g_id = item_id_com[k]
        #
        if g_id not in collision_groups:
            collision_groups[g_id] = set()
        for v in v_ls:
            collision_groups[g_id].add(item_id_com[v])

    return collision_groups



from kevin_toolbox.machine_learning.dataset.face.verification import Factory
from . import __by_block


def get_executor_ls_by_block(**kwargs):
    """
        通过调用 verification.Factory 中的 generate_by_block() 函数，
            生成一系列的执行器 executor_ls，
            每个执行器在被 executor() 调用后都将返回一个数据集

        参数：
            mode:                   构造哪种迭代器，现在支持以下模式：
                                        "all":  对整个特征矩阵进行构造
                                        "triangle": 对特征矩阵的上三角进行构造
            factory:                verification.Factory 实例

        设定数据集大小：
            chunk_step:             每个分块的大小（根据实际内存容量来选择）
                                        限制了每次返回的数据集大小。
            upper_bound_of_dataset_size:        每次返回的数据集的大小的上界（根据实际内存容量来选择）
                                        当给定有 upper_bound_of_dataset_size 时，
                                        程序将可以通过 cal_chunk_step() 来计算得到 chunk_step。
                                        该参数是间接设置 chunk_step 的方式，但相较而言更加直观。
            （建议使用 upper_bound_of_dataset_size 而不使用 chunk_step。两者同时设置时，将以两者中最小的为限制）

        输入到 Factory.generate_by_block 中：
            need_to_generate:       需要生成的字段
                                        （参见 Factory.generate_by_block() 中的介绍）
            include_diagonal:       是否包含对角线
    """
    # 默认参数
    paras = {
        "mode": "triangle",
        "factory": None,
        "chunk_step": None,
        "upper_bound_of_dataset_size": None,
        "need_to_generate": None,
        "include_diagonal": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert paras["mode"] in ["all", "triangle"]
    assert isinstance(paras["factory"], (Factory,)), \
        Exception(f"Error: The type of input factory should be {Factory}, but get a {type(paras['factory'])}!")

    temp_ls = []
    if paras["upper_bound_of_dataset_size"] is not None:
        temp_ls.append(cal_chunk_step(paras["upper_bound_of_dataset_size"]))
    if paras["chunk_step"] is not None:
        temp_ls.append(paras["chunk_step"])
    assert len(temp_ls) > 0
    paras["chunk_step"] = min(temp_ls)

    #
    if paras["mode"] == "all":
        executor_ls, size_ls = __by_block.of_all(**paras)
    else:
        executor_ls, size_ls = __by_block.of_triangle(**paras)
    return executor_ls, size_ls


def cal_chunk_step(upper_bound):
    """
        根据数据集大小的限制，计算每个数据块的边长上界。
        返回值满足:
            chunk_step = argmax( chunk_step * (chunk_step + 1) <= upper_bound )
    """
    # block的宽
    return int(((1 + 4 * upper_bound) ** 0.5 - 1) / 2)



import numpy as np


def detect_collision_inside_boxes(**kwargs):
    """
        基于分离轴定理，对 box（属于凸多边形），进行碰撞检测
            特点：
                - 时复杂度为 O( N*log(N) + M ) 其中 N 表示 box 的数量，M 表示发生碰撞的 pairs 数量。
                - 不需要像AABB包围盒和四叉树那样依赖树结构。
                - 可以配合 for_boxes.convert_from_coord_to_grid_index() 将 boxes 映射到格点阵列内，
                    从而实现多阶段碰撞检测的 Broad-Phase 和 Narrow-Phase。
            注意：
                - 我们将接触但不重合的情况也视为是碰撞。

        基本流程：
            1. 计算各个轴的碰撞概率
            2. 选取碰撞概率最小的轴开始进行 aixes_check
                （aixes_check 是基于分离轴定理，使用 Sort and Sweep 的方式进行的碰撞粗检测）
            3. 比较后续进行 aixes_check 和 fine_check 的时间成本，选择成本最低的方式进行迭代
                （fine_check 是对前面碰撞粗检测得到的潜在碰撞 box pairs 进行逐一精细准确的碰撞检测）

        参数：
            boxes:          <3 axis np.array> 需要检测的 box
                                shape [batch_size, 2, dimensions]，各个维度的意义为：
                                    batch_size： 有多少个 box
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
            complexity_correction_factor_for_aixes_check:   <float/integer> 进行 aixes_check 的复杂度修正系数
                                通过设置该系数，可以调整 aixes_check 与 fine_check 的复杂度比例。
                                - 该系数越大，计算得到 aixes_check 的复杂度越高，程序对于进行 fine_check 的偏好越大。
                                - 该系数越小，对进行 aixes_check 的偏好越大。
                                系数默认为 1.0，建议根据不同设备的实际情况（最好进行测试比较）进行调整。
            duplicate_records:  <boolean> 是否在输出的 collision_groups 的每个 box_id 下都记录一次碰撞。
                                    默认为 False，此时对于每个碰撞对，只会在其中一个 box_id 下的 set 中记录一次。
                                        至于碰撞对的具体分配方式则是随机的，不应作为后续流程依仗的特征。
                                    当设置为 True，则会重复记录。
        输出：
            collision_groups:   <dict of integers set> 检出的碰撞对。
                                其中的第 i 个 set 记录了 box_id==i 的 box 与其他哪些 box 存在碰撞
                                    比如：  collision_groups[0] = {1, 2} 表示0号 box 与1、2号 box 发生了碰撞
    """
    # 默认参数
    paras = {
        # 必要参数
        "boxes": None,
        #
        "complexity_correction_factor_for_aixes_check": 1.0,
        # setting for output
        "duplicate_records": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    paras["boxes"] = np.asarray(paras["boxes"])
    assert paras["boxes"].ndim == 3 and paras["boxes"].shape[1] == 2
    boxes = np.sort(paras["boxes"], axis=1)  # 对两个端点进行标准化
    #
    assert isinstance(paras["complexity_correction_factor_for_aixes_check"], (int, float,))
    correction_factor = paras["complexity_correction_factor_for_aixes_check"]

    # 1. 计算各个轴的碰撞概率
    #   - 通过指标 box在该轴上投影长度的均值/box的中点的均方差 估计各个轴上检出碰撞的概率，该指标越小，碰撞概率小。
    edge_lens = boxes[:, 1, :] - boxes[:, 0, :]
    indicators = np.mean(edge_lens, axis=0) / (np.std(edge_lens, axis=0) + 1e-10)  # 指标
    indicator_and_dims = [(indicator, dim) for dim, indicator in enumerate(indicators)]
    indicator_and_dims.sort(key=lambda x: x[0])  # [(indicator, dim)] sorted by indicator

    time_cost_for_aixes_check = len(boxes) * np.log(len(boxes)) * 2 * correction_factor

    # 2. 选取碰撞概率最小的轴开始进行 aixes_check
    #   - 从碰撞稀疏的轴开始进行碰撞检测，能够缩小每个碰撞group的size，从而降低在此基础上进行 fine_check 的算法复杂度。
    #   为及时从 aixes_check 切换到 fine_check 做好准备。
    last_collision_groups = None
    collision_groups = None
    for step, (_, dim) in enumerate(indicator_and_dims):
        # 2.1 获取各个box在该轴上的投影
        ticks = dict()
        for box_id in range(len(boxes)):
            (p_0, p_1) = boxes[box_id]
            beg, end = p_0[dim], p_1[dim]
            for tick in [beg, end]:
                if tick not in ticks:
                    ticks[tick] = list()
                ticks[tick].append(box_id)
        sort_ticks = sorted(ticks.items(), key=lambda x: x[0])

        # 2.2 检测潜在的碰撞 pairs
        #   - 只要轴上的投影有重合，就视为潜在的碰撞。
        meet_beg_ids = set()
        # collision_groups 中的第 i 个列表记录了 box_id==i 在轴上的投影与其他哪些 box 存在碰撞
        # 例如 collision_groups[5] = {6, 1} 表示5号框与6、1号框发生了碰撞
        collision_groups = dict()
        for k, v_ls in sort_ticks:
            # 将 meet_end_id_ls 的生成与 collision_groups 的生成分离开来是为了适应 接触但不重合的情况 和 box 体积为0的情况
            meet_end_id_ls = list()
            for v in v_ls:
                if v in meet_beg_ids:
                    # meet the end
                    meet_end_id_ls.append(v)
                else:
                    # new beg
                    meet_beg_ids.add(v)
            # 保证每次碰撞只被记录最多一次
            for v in meet_end_id_ls:
                meet_beg_ids.remove(v)
                collision_groups[v] = meet_beg_ids.copy()

        # 2.3 根据本次检测结果，剔除之前检测中的疑似碰撞 pairs
        if step == 0:
            last_collision_groups = collision_groups
        else:
            # 3. 比较后续进行 aixes_check 和 fine_check 的时间成本，选择成本最低的方式进行迭代
            time_cost_for_fine_check = sum([len(g) for g in collision_groups.values()])
            if time_cost_for_aixes_check * (boxes.shape[-1] - 1 - step) > time_cost_for_fine_check:
                # 进行 fine_check 的时间成本更低
                # 直接进行一次 fine_check 就可以得到最终结果，跳出循环
                for i, j_set in collision_groups.items():
                    true_collisions = set()
                    for j in j_set:
                        # intersection
                        beg = np.maximum(boxes[i][0], boxes[j][0])
                        end = np.minimum(boxes[i][1], boxes[j][1])
                        if (end - beg >= 0).all():  # 对于接触但不重合的情况，也视为碰撞
                            true_collisions.add(j)
                    collision_groups[i] = true_collisions
                break
            else:
                # 进行 aixes_check 的时间成本更低
                # 合并之前的 aixes_check 结果 last_collision_groups，然后继续下次循环
                for i, j_set in collision_groups.items():
                    for j in j_set.copy():
                        if j in last_collision_groups[i] or i in last_collision_groups[j]:
                            # 两边都检测出碰撞
                            pass
                        else:
                            j_set.remove(j)
                last_collision_groups = collision_groups

    #
    if paras["duplicate_records"]:
        for i, j_set in collision_groups.items():
            for j in j_set:
                collision_groups[j].add(i)
    for i in list(collision_groups.keys()):
        if len(collision_groups[i]) == 0:
            collision_groups.pop(i)

    return collision_groups
